import json

from amscrot.model import Node
from amscrot.util.utils import get_logger
from amscrot.util.constants import Constants
from .cloudlab_constants import *
from .cloudlab_exceptions import CloudlabException
from .cloudlab_provider import CloudlabProvider

logger = get_logger()


class CloudlabNode(Node):
    def __init__(self, *, label, name: str, provider: CloudlabProvider, network):
        super().__init__(label=label, name=name, image="", site="", flavor="")
        self._provider = provider
        self._net = network
        self.dataplane_ipv4 = None
        self.dataplane_ipv6 = None
        self.keyfile = provider.private_key_file_location
        self.user = provider.user

    @property
    def provider(self):
        return self._provider

    @property
    def net(self):
        return self._net

    def create(self):
        import emulab_sslxmlrpc.client.api as api

        server = self.provider.rpc_server()
        exp_params = self.provider.experiment_params(self.net.name)
        exitval, response = api.experimentStatus(server, exp_params).apply()

        if exitval:
            raise CloudlabException(exitval=exitval, response=response)

        status = json.loads(response.value)
        logger.info(f"STATUS: {json.dumps(status, indent=2)}")

        if status['status'] != 'ready':
            raise CloudlabException(message=f"Found experiment with status={status['status']}")

        node_uris = list(status[AGGREGATE_STATUS].keys())
        node_uri = [uri for uri in  node_uris if 'stitch' not in uri][0]
        idx = int(self.name[self.name.rindex('-') + 1:])
        node_info = status[AGGREGATE_STATUS][node_uri][NODES]["node" + str(idx)]
        logger.info(f"NODE_INFO {self.name}: {node_info}")

        self.mgmt_ip = node_info[IPV4]
        self.host = node_info[IPV4]
        exitval, response = api.experimentManifests(server, exp_params).apply()

        if exitval:
            raise CloudlabException(exitval=exitval, response=response)

        status = json.loads(response.value)

        logger.debug(json.dumps(status, indent=2))

        import xmltodict

        data_dict = xmltodict.parse(list(status.values())[0])
        logger.debug(f"RSPEC: {json.dumps(data_dict['rspec'], indent=2)}")
        nodes = data_dict['rspec']['node']
        nodes = [n for n in nodes if self._net.cluster == n['@component_manager_id']]
        n = nodes[idx]

        if n['interface']['ip']['@type'] == 'ipv4':
            self.dataplane_ipv4 = n['interface']['ip']['@address']
        else:
            self.dataplane_ipv6 = n['interface']['ip']['@address']

        self.site = n['jacks:site']['@id']

    def delete(self):
        pass

    def get_reservation_state(self) -> str:
        pass

    def get_reservation_id(self) -> str:
        pass

    def add_route(self, subnet, gateway):
        pass

    def get_dataplane_address(self, network=None, interface=None, af=Constants.IPv4):
        if af == Constants.IPv4:
            return self.dataplane_ipv4
        elif af == Constants.IPv6:
            return self.dataplane_ipv6
        else:
            return None



CHI_USER = "user"
CHI_PASSWORD = "password"
CHI_AUTH_URL = "auth_url"
CHI_TACC = "tacc"
CHI_UC = "uc"
CHI_KVM = "kvm"
CHI_EDGE = "edge"
CHI_CLIENT_ID = "client_id"
CHI_PROJECT_NAME = "project_name"
CHI_PROJECT_ID = "project_id"
CHI_RANDOM = "CHI.RANDOM"
CHI_KEY_PAIR = "key_pair"
CHI_SLICE_PRIVATE_KEY_LOCATION = "slice-private-key-location"
CHI_SLICE_PUBLIC_KEY_LOCATION = "slice-public-key-location"

CHI_CONF_ATTRS = [CHI_USER,
                  CHI_PASSWORD,
                  CHI_KEY_PAIR,
                  CHI_PROJECT_NAME,
                  CHI_PROJECT_ID,
                  CHI_SLICE_PRIVATE_KEY_LOCATION,
                  CHI_SLICE_PUBLIC_KEY_LOCATION]

DEFAULT_NETWORK = "sharednet1"
DEFAULT_NETWORKS = [DEFAULT_NETWORK, "sharedwan1", "containernet1"]
DEFAULT_IMAGE = "CC-Ubuntu20.04"
DEFAULT_FLAVOR = "m1.medium"

DEFAULT_AUTH_URLS = dict(
    tacc='https://chi.tacc.chameleoncloud.org:5000/v3',
    uc='https://chi.uc.chameleoncloud.org:5000/v3',
    kvm='https://kvm.tacc.chameleoncloud.org:5000/v3',
    edge='https://chi.edge.chameleoncloud.org:5000/v3'
)

DEFAULT_CLIENT_IDS = dict(
    tacc='keystone-tacc-prod',
    uc='keystone-uc-prod',
    kvm='keystone-kvm-prod',
    edge='keystone-edge-prod'
)

DEFAULT_PROJECT_IDS = dict(
    tacc='a400724e818d40cbba1a5c6b5e714462',
    uc='ae76673270164b048b59d3bd30676721',
    kvm='a400724e818d40cbba1a5c6b5e714462',
    edge=None
)

DEFAULT_DISCOVERY_URL = "https://auth.chameleoncloud.org/auth/realms/chameleon/.well-known/openid-configuration"

INCLUDE_ROUTER = True 



import json

from amscrot.model import Network
from amscrot.util.constants import Constants
from amscrot.util.utils import get_logger
from .cloudlab_constants import *
from .cloudlab_exceptions import CloudlabException
from .cloudlab_provider import CloudlabProvider

logger = get_logger()


class CloudNetwork(Network):
    def __init__(self, *, label, name: str, provider: CloudlabProvider,
                 stitch_info=None, profile: str, interfaces, layer3, cluster):
        super().__init__(label=label, name=name, site="")
        self.profile = profile
        self._provider = provider
        self.stitch_info = stitch_info
        self.interface = interfaces or []
        self.layer3 = layer3
        self.cluster = cluster
        self.stich_node_ip = None
        self.stich_site = None

    @property
    def provider(self):
        return self._provider

    @property
    def project(self):
        return self.provider.config[CLOUDLAB_PROJECT]

    def params(self):
        params = {
            "profile": f"{self.project},{self.profile}",
            "proj": self.project,
            "name": self.name,
            "asjson": True,
        }

        return params

    def create(self):
        import emulab_sslxmlrpc.client.api as api
        import emulab_sslxmlrpc.xmlrpc as xmlrpc

        server = self.provider.rpc_server()
        exp_params = self.provider.experiment_params(self.name)
        exitval, response = api.experimentStatus(server, exp_params).apply()

        if exitval == xmlrpc.RESPONSE_SEARCHFAILED:
            logger.debug(f"Network {self.name} not found, creating...")
            params = self.params()
            bindings = {}

            if self.interface:
                vlan = self.interface[0]['vlan']
                bindings['vlan'] = str(vlan)

            bindings['cluster'] = self.cluster
            # noinspection PyUnresolvedReferences
            nodes = [n for n in self.provider.nodes if n.net == self]
            bindings['node_count'] = str(len(nodes))
            subnet = self.layer3.attributes[Constants.RES_SUBNET]
            bindings['ip_subnet'] = str(subnet)
            ip_start = self.layer3.attributes[Constants.RES_LAYER3_DHCP_START]
            bindings['ip_start'] = str(ip_start)
            params['bindings'] = json.dumps(bindings)
            logger.info(f"Network {self.name} not found, creating... {params}")
            exitval, response = api.startExperiment(server, params).apply()

            if exitval:
                raise CloudlabException(exitval=exitval, response=response)

        elif exitval == xmlrpc.RESPONSE_SUCCESS or exitval == xmlrpc.RESPONSE_ALREADYEXISTS:
            logger.info("Network already exists, checking status")
        else:
            raise CloudlabException(exitval=exitval, response=response)

    def wait_for_create(self):
        import time
        import emulab_sslxmlrpc.client.api as api

        server = self.provider.rpc_server()
        exp_params = self.provider.experiment_params(self.name)
        attempt = -1

        for attempt in range(CLOUDLAB_RETRY):
            exitval, response = api.experimentStatus(server, exp_params).apply()

            # sometimes the response is not what we expect (network glitch). We keep checking status ...
            if response and hasattr(response, "value"):
                if exitval:
                    code = response.value

                    if code == api.GENIRESPONSE_REFUSED or code == api.GENIRESPONSE_NETWORK_ERROR:
                        logger.debug("Server is offline, waiting for a bit")
                    elif code == api.GENIRESPONSE_BUSY:
                        logger.debug("Experiment is busy, waiting for a bit")
                    elif code == api.GENIRESPONSE_SEARCHFAILED:
                        raise CloudlabException(message="Experiment is gone", exitval=exitval, response=response)
                    else:
                        raise CloudlabException(exitval=exitval, response=response)
                else:
                    status = json.loads(response.value)
                    logger.info(json.dumps(status, indent=2))

                    if status["status"] == "failed":
                        raise CloudlabException(exitval=exitval, response=response,
                                                message="Experiment failed to instantiate")
                    elif status["status"] == "ready":
                        if "execute_status" not in status:
                            logger.info("No execute service to wait for!")
                            break

                        total = status["execute_status"]["total"]
                        finished = status["execute_status"]["finished"]

                        if total != finished:
                            logger.info("Still waiting for execute service to finish")
                            continue

                        logger.info("Execute services have finished")
                        break

            if response and hasattr(response, "value"):
                logger.info(f"Still waiting for experiment to be ready exitval={exitval}:{response.value}")
            else:
                logger.warning(f"Still waiting for experiment to be ready exitval={exitval}:{response}")

            time.sleep(CLOUDLAB_SLEEP_TIME)

        if attempt == CLOUDLAB_RETRY:
            raise CloudlabException(message="Please Apply Again. Giving up on waiting for experiment ...")

        exitval, response = api.experimentManifests(server, exp_params).apply()

        if exitval:
            raise CloudlabException(exitval=exitval, response=response)

        status = json.loads(response.value)
        logger.debug(json.dumps(status, indent=2))
        logger.info(f"STATUS_KEYS={status.keys()}")

        import xmltodict

        data_dict = xmltodict.parse(next(iter(status.values())))
        logger.info(f"RSPEC: {json.dumps(data_dict['rspec'], indent=2)}")
        link = data_dict['rspec']['link']

        temp = dict(id=self.label, vlan=link['@vlantag'])
        temp.update(self.stitch_info.stitch_port['peer'])
        temp['provider'] = self.stitch_info.stitch_port['provider']
        self.interface = [temp]

        # noinspection PyUnresolvedReferences
        if not [n for n in self.provider.nodes if n.net == self]:
            return

        all_nodes = data_dict['rspec']['node']
        nodes = [n for n in all_nodes if 'stitch' not in ['@component_manager_id']]
        n = nodes[0]
        self.stich_node_ip = n['interface']['ip']['@address']
        self.stich_site = n['jacks:site']['@id']
        nodes = [n for n in all_nodes if 'stitch' not in n['@component_manager_id']]
        self.site = nodes[0]['jacks:site']['@id']

    def delete(self):
        import emulab_sslxmlrpc.client.api as api
        import emulab_sslxmlrpc.xmlrpc as xmlrpc
        import time

        server = self.provider.rpc_server()
        exp_params = self.provider.experiment_params(self.name)
        exitval, response = api.experimentStatus(server, exp_params).apply()

        if exitval == xmlrpc.RESPONSE_SEARCHFAILED:
            return

        exitval, response = api.terminateExperiment(server, exp_params).apply()

        if exitval == xmlrpc.RESPONSE_SUCCESS:
            while True:
                exitval, response = api.experimentStatus(server, exp_params).apply()

                if exitval == xmlrpc.RESPONSE_SEARCHFAILED:
                    break

                logger.info("Still waiting for experiment to be terminated")
                time.sleep(CLOUDLAB_SLEEP_TIME)

        if exitval != xmlrpc.RESPONSE_SEARCHFAILED:
            raise CloudlabException(exitval=exitval, response=response)



from amscrot.model import Network
from amscrot.util.utils import get_logger
from . import sense_utils
from .sense_constants import SERVICE_INSTANCE_KEYS
from .sense_exceptions import SenseException
import json

logger = get_logger()


class SenseNetwork(Network):
    def __init__(self, *, label, name: str, bandwidth, profile, layer3, peering, interfaces, saved_interfaces=None):
        super().__init__(label=label, name=name, site="")
        self.profile = profile
        self.layer3 = layer3
        self.peering = peering
        self.interface = interfaces or list()
        self.bandwidth = bandwidth
        self.dtn = []
        self.id = ''
        self.referenceUUID = self.id
        self.intents = []
        self._saved_interfaces = saved_interfaces or list()

    # CREATE - COMPILED, CREATE - COMMITTING, CREATE - COMMITTED, CREATE - READY
    def create(self):
        si_uuid = sense_utils.find_instance_by_alias(alias=self.name)

        if not si_uuid:
            logger.debug(f"Creating {self.name}")
            self._saved_interfaces = list()
            si_uuid, status = sense_utils.create_instance(profile=self.profile, bandwidth=self.bandwidth,
                                                          alias=self.name,
                                                          layer3=self.layer3, peering=self.peering,
                                                          interfaces=self.interface)
        else:
            logger.debug(f"Found {self.name} {si_uuid}")
            assert si_uuid
            status = sense_utils.instance_get_status(si_uuid=si_uuid)
            logger.info(f"Found existing {self.name} {si_uuid} with status={status}")

        if 'FAILED' in status:
            raise SenseException(f"Found instance {si_uuid} with status={status}")

        if 'CREATE - READY' not in status:
            logger.debug(f"Provisioning {self.name}")
            sense_utils.instance_operate(si_uuid=si_uuid)

        # statuses = ('CREATE - READY', 'REINSTATE - READY', 'CREATE - COMMITTED', 'REINSTATE - COMMITTED')
        statuses = ('CREATE - READY', 'REINSTATE - READY')
        status = sense_utils.wait_for_instance_operate(si_uuid=si_uuid, statuses=statuses)

        if status not in statuses:
            raise SenseException(f"Creation failed for {si_uuid} {status}")

        logger.debug(f"Retrieving details {self.name} {status}")
        instance_dict = sense_utils.service_instance_details(si_uuid=si_uuid, alias=self.name)
        logger.info(f"Retrieved details {self.name} {status}: \n{json.dumps(instance_dict, indent=2)}")

        for key in SERVICE_INSTANCE_KEYS:
            self.__setattr__(key, instance_dict.get(key))

        self.id = self.referenceUUID

        if (self.intents[0]['json']['service'] != 'vcn' or
                "GCP" not in self.intents[0]['json']['data']['gateways'][0]['type'].upper()):
            return

        for saved_interface in self._saved_interfaces:
            if saved_interface.get('kind') == 'gcp_pairing_key':
                self.interface.append(saved_interface)
                return

        from .sense_constants import SENSE_RETRY

        for attempt in range(SENSE_RETRY):
            try:
                template_file = 'gcp-pairing-key-template.json'
                details = sense_utils.manifest_create(si_uuid=si_uuid, template_file=template_file)
                pairing_key = details.get('Pairing Key', "?pairing_key?")

                if '?pairing_key?' not in pairing_key:
                    self.interface.append(dict(id=pairing_key, provider="sense", info=status, kind='gcp_pairing_key'))
                    break

            except Exception as e:
                if status in ('CREATE - READY', 'REINSTATE - READY'):
                    raise SenseException(f"not able to get pairing key ...status={status}:{e}")

            logger.warning(f"Waiting on pairing key: going to sleep attempt={attempt}:status={status}")
            import time

            time.sleep(2)
            status = sense_utils.instance_get_status(si_uuid=si_uuid)

    def delete(self):
        from . import sense_utils

        si_uuid = sense_utils.find_instance_by_alias(alias=self.name)

        logger.debug(f"Deleting {self.name} {si_uuid}")

        if si_uuid:
            sense_utils.delete_instance(si_uuid=si_uuid)
            logger.debug(f"Deleted {self.name} {si_uuid}")



from amscrot.model import Network
from amscrot.util.utils import get_logger, Constants
from . import gcp_utils
from .gcp_exceptions import GcpException
from .gcp_provider import GcpProvider

logger = get_logger()


class GcpNetwork(Network):
    def __init__(self, *, label, name: str, provider: GcpProvider, layer3, peering, stitch_port):
        super().__init__(label=label, name=name, site="")
        self._provider = provider
        self.layer3 = layer3
        self.peering = peering
        self.stitch_port = stitch_port
        self.interface = []

    def create(self):
        project = self._provider.project
        service_key_path = self._provider.service_key_path
        vpc = self.peering.attributes.get(Constants.RES_CLOUD_VPC)

        if not vpc:
            raise GcpException(f"Must supply Vpc using peering config and {Constants.RES_CLOUD_VPC}")

        vpc_details = gcp_utils.find_vpc(service_key_path=service_key_path,
                                         project=project,
                                         vpc=vpc)

        if not vpc_details:
            raise GcpException(f"Vpc {vpc} not found")

        logger.info(f"vpc_details={vpc_details}")

        region = self.peering.attributes.get(Constants.RES_CLOUD_REGION)

        if not region:
            region = self.stitch_port['peer'].get(Constants.STITCH_PORT_REGION)

        if not region:
            raise GcpException(f"Missing cloud region")

        router_name = f'{self.name}-router'
        router = gcp_utils.find_router(service_key_path=service_key_path,
                                       project=project,
                                       region=region,
                                       router_name=router_name)

        if not router:
            google_asn = self.peering.attributes.get(Constants.RES_REMOTE_ASN)

            if isinstance(google_asn, str):
                google_asn = int(google_asn)

            vpc = self.peering.attributes.get(Constants.RES_CLOUD_VPC)
            gcp_utils.create_router(service_key_path=service_key_path,
                                    project=project,
                                    region=region,
                                    router_name=router_name,
                                    vpc=vpc,
                                    bgp_asn=google_asn)

        attachment_name = f'{self.name}-vlan-attachment'
        attachment = gcp_utils.find_interconnect_attachment(service_key_path=service_key_path,
                                                            project=project,
                                                            region=region,
                                                            attachment_name=attachment_name)

        if not attachment:
            mtu = self.peering.attributes.get(Constants.RES_CLOUD_MTU, 1460)
            gcp_utils.create_interconnect_attachment(service_key_path=service_key_path,
                                                     project=project,
                                                     region=region,
                                                     mtu=int(mtu),
                                                     router_name=router_name,
                                                     attachment_name=attachment_name)
            attachment = gcp_utils.find_interconnect_attachment(service_key_path=service_key_path,
                                                                project=project,
                                                                region=region,
                                                                attachment_name=attachment_name)

        # set the MD5 authentication
        assert Constants.RES_SECURITY in self.peering.attributes
        bgp_key = self.peering.attributes[Constants.RES_SECURITY]

        gcp_utils.patch_router(service_key_path=service_key_path,
                               project=project,
                               region=region,
                               router_name=router_name,
                               bgp_key=bgp_key)

        logger.info(f"attachment_details={attachment}")
        self.interface.append(dict(id=attachment.pairing_key, provider=self._provider.type))

    def delete(self):
        project = self._provider.project
        service_key_path = self._provider.service_key_path
        region = self.peering.attributes.get(Constants.RES_CLOUD_REGION)

        if not region:
            region = self.stitch_port['peer'].get(Constants.STITCH_PORT_REGION)

        if not region:
            raise GcpException(f"Missing cloud region")

        attachment_name = f'{self.name}-vlan-attachment'
        attachment = gcp_utils.find_interconnect_attachment(service_key_path=service_key_path,
                                                            project=project,
                                                            region=region,
                                                            attachment_name=attachment_name)

        if attachment:
            gcp_utils.delete_interconnect_vlan_attachment(service_key_path=service_key_path,
                                                          project=project,
                                                          region=region,
                                                          attachment_name=attachment_name)

        router_name = f'{self.name}-router'
        router = gcp_utils.find_router(service_key_path=service_key_path,
                                       project=project,
                                       region=region,
                                       router_name=router_name)

        if router:
            gcp_utils.delete_router(service_key_path=service_key_path,
                                    project=project,
                                    region=region,
                                    router_name=router_name)



from enum import Enum
from typing import List, Dict, Any, Optional, Union
from amscrot.serviceclient import ServiceClient

class JobServiceType(str, Enum):
    REALTIME = "REALTIME"
    BATCH = "BATCH"
    INTERACTIVE = "INTERACTIVE"

class JobType(str, Enum):
    COMPUTE = "COMPUTE"
    NETWORK = "NETWORK"
    STORAGE = "STORAGE"
    DATA = "DATA"
    INSTRUMENT = "INSTRUMENT"

class JobState(str, Enum):
    # amscrot lifecycle states (not from esnet-iri)
    INIT      = "INIT"       # Job object created, not yet planned
    PLANNED   = "PLANNED"    # Job validated/planned successfully
    PENDING   = "PENDING"    # Submitted, awaiting provider scheduling
    UNKNOWN   = "UNKNOWN"    # State cannot be determined

    # States aligned with esnet-iri JobState
    NEW       = "NEW"        # Job created at the provider
    QUEUED    = "QUEUED"     # Job queued at the provider
    ACTIVE    = "ACTIVE"     # Job is actively running
    COMPLETED = "COMPLETED"  # Job completed successfully
    FAILED    = "FAILED"     # Job failed
    CANCELED  = "CANCELED"   # Job was canceled/destroyed

class JobStatus:
    """Result of a ServiceClient.status() call.

    Wraps a normalized status with the serialized provider-specific response.
    """
    def __init__(
        self,
        state: str = "UNKNOWN",
        message: Optional[str] = None,
        exit_code: Optional[int] = None,
        job_id: Optional[str] = None,
        resource_id: Optional[str] = None,
        provider_status: Optional[Dict[str, Any]] = None,
    ):
        self.state = state
        self.message = message
        self.exit_code = exit_code
        self.job_id = job_id
        self.resource_id = resource_id
        self.provider_status = provider_status

    def to_dict(self) -> Dict[str, Any]:
        d: Dict[str, Any] = {"state": self.state}
        if self.message is not None:
            d["message"] = self.message
        if self.exit_code is not None:
            d["exit_code"] = self.exit_code
        if self.job_id is not None:
            d["job_id"] = self.job_id
        if self.resource_id is not None:
            d["resource_id"] = self.resource_id
        if self.provider_status is not None:
            d["provider_status"] = self.provider_status
        return d

    def __repr__(self):
        parts = [f"state={self.state!r}"]
        if self.message:
            parts.append(f"message={self.message!r}")
        if self.exit_code is not None:
            parts.append(f"exit_code={self.exit_code}")
        if self.job_id:
            parts.append(f"job_id={self.job_id!r}")
        return f"<JobStatus {' '.join(parts)}>"

class JobSpec:
    def __init__(self, resources: Dict = None, image: str = None, executable: List[str] = None, attributes: Dict = None):
        self.resources = resources or {}
        self.image = image
        self.executable = executable or []
        self.attributes = attributes or {}

    def to_dict(self) -> Dict:
        return {
            'resources': self.resources,
            'image': self.image,
            'executable': self.executable,
            'attributes': self.attributes
        }

class Job:
    def __init__(
        self,
        name: str,
        type: Union[JobType, str],
        service_type: Union[JobServiceType, str],
        service_client: Optional[ServiceClient] = None,
        job_spec: Optional[JobSpec] = None,
        dependency: Optional['Job'] = None,
        preferences: Dict = None,
        inputs: Dict = None,
        outputs: Dict = None
    ):
        self.name = name
        self.type = JobType(type) if isinstance(type, str) else type
        self.service_type = JobServiceType(service_type) if isinstance(service_type, str) else service_type
        
        self.status = JobState.INIT
        self.service_client = service_client
        self.job_spec = job_spec or JobSpec()
        self.dependency = dependency
        self.preferences = preferences or {}
        self.inputs = inputs or {}
        self.outputs = outputs or {}
        self.local_files: Dict[str, str] = {}  # populated by fetch_output_files

    def to_config(self) -> Dict:
        config = {
            'name': self.name,
            'type': self.type.value if hasattr(self.type, 'value') else self.type,
            'service_type': self.service_type.value if hasattr(self.service_type, 'value') else self.service_type,
            'status': self.status.value if hasattr(self.status, 'value') else self.status,
            'spec': self.job_spec.to_dict(),
            'preferences': self.preferences,
            'inputs': self.inputs,
            'outputs': self.outputs
        }
        if self.service_client:
            config['service_client'] = self.service_client.name # Reference by name
        if self.dependency:
            config['dependency'] = self.dependency.name # Reference by name
            
        return {self.name: config}

    def set_status(self, status: Union[JobState, str]):
        self.status = JobState(status) if isinstance(status, str) else status

    def __repr__(self):
        return f"<Job name={self.name} type={self.type} status={self.status}>"



from ipaddress import IPv4Address
from typing import List, Dict, Union

from fabrictestbed_extensions.fablib.network_service import NetworkService
from fabrictestbed_extensions.fablib.slice import Slice

from amscrot.model import Network
from amscrot.policy.policy_helper import get_stitch_port_for_provider
from amscrot.util.constants import Constants
from amscrot.util.utils import get_logger
from .fabric_provider import FabricProvider
from ...util.config_models import Config

logger = get_logger()


class FabricNetwork(Network):
    def __init__(self, *, label, delegate: NetworkService, layer3: Config,
                 peering: Union[List[Config], Config], peer_layer3: Config):
        self.name = delegate.get_name()
        self.site = delegate.get_site()
        super().__init__(label=label, name=self.name, site=self.site)
        self._delegate = delegate
        self.site = delegate.get_site()
        self.slice_name = self._delegate.get_slice().get_name()
        self.type = str(delegate.get_type())

        self.layer3 = layer3
        self.peering = peering
        self.peer_layer3 = peer_layer3
        ns = self._delegate.get_fim_network_service()
        self.interface = []

        if self.peering and isinstance(self.peering, list):
            for peering in self.peering:
                if Constants.RES_CLOUD_ACCOUNT in peering.attributes:
                    self.interface.append(dict(id=self.slice_name, provider="fabric"))
        elif self.peering and Constants.RES_CLOUD_ACCOUNT in self.peering.attributes:
            self.interface.append(dict(id=self.slice_name, provider="fabric"))

        for key, iface in ns.interfaces.items():
            if hasattr(iface.labels, "vlan") and iface.labels.vlan:
                self.interface.append(dict(id=key, vlan=iface.labels.vlan))

        self.id = self._delegate.get_reservation_id()
        self.state = self._delegate.get_reservation_state()

    @property
    def subnet(self):
        return self.layer3.attributes.get(Constants.RES_SUBNET) if self.layer3 else None

    @property
    def gateway(self):
        return self.layer3.attributes.get(Constants.RES_NET_GATEWAY) if self.layer3 else None

    @property
    def delegate(self):
        return self._delegate

    def available_ips(self) -> list:
        available_ips = []

        if self.layer3:
            ip_start = self.layer3.attributes.get(Constants.RES_LAYER3_DHCP_START)
            ip_end = self.layer3.attributes.get(Constants.RES_LAYER3_DHCP_END)

            if ip_start and ip_end:
                pool_start = int(IPv4Address(ip_start))
                pool_end = int(IPv4Address(ip_end))

                for ip_int in range(pool_start + 1, pool_end + 1):
                    available_ips.append(IPv4Address(ip_int))

        return available_ips

    def get_reservation_id(self):
        self._delegate.get_reservation_id()

    def get_site(self):
        return self._delegate.get_site()


class NetworkBuilder:
    def __init__(self, label, provider: FabricProvider, slice_object: Slice, name, resource: dict):
        from amscrot.policy.policy_helper import StitchInfo

        self.provider = provider
        self.slice_object = slice_object
        self.stitch_infos: List[StitchInfo] = resource.get(Constants.RES_STITCH_INFO)
        self.stitch_port = get_stitch_port_for_provider(resource=resource, provider=provider.type)
        self.site = resource.get(Constants.RES_SITE)
        self.facility_cloud_port_interfaces = []
        self.facility_non_cloud_port_interfaces = []
        self.net_name = name
        self.layer3 = resource.get(Constants.RES_LAYER3)
        self.peering: Union[List[Config], Config] = resource.get(Constants.RES_PEERING)
        self.peer_layer3 = resource.get(Constants.RES_PEER_LAYER3)
        self.stitching_nets = []
        self.label = label
        self.net: Union[NetworkService, None] = None
        self.bw = resource.get(Constants.RES_BW)
        self.discovered_stitch_infos: List[Dict] = []
        self.sites = set()

        import amscrot.provider.api.dependency_util as util

        if util.has_resolved_external_dependencies(resource=resource, attribute=Constants.RES_STITCH_INTERFACE):
            self.stitching_nets = util.get_values_for_dependency(resource=resource,
                                                                 attribute=Constants.RES_STITCH_INTERFACE)

            for stitching_net in self.stitching_nets:
                interface = stitching_net.interface
                assert isinstance(interface, list)
                interface = next(iter([i for i in interface if isinstance(i, dict) and 'provider' in i]))
                logger.info(f'Network {self.net_name} got stitching interface {interface} from {stitching_net.label}')
                discovered_stitch_info = dict()
                discovered_stitch_info.update(interface)
                self.discovered_stitch_infos.append(discovered_stitch_info)

        if self.peering and not isinstance(self.peering, list):
            self.peering = [self.peering]

        if self.stitch_port and not isinstance(self.stitch_port, list):
            self.stitch_port = [self.stitch_port]

        if self.peering:
            from .plugins import Plugins

            Plugins.load()

    @staticmethod
    def _get_stitch_info_index(peering, stitch_ports):
        for idx, stitch_port in enumerate(stitch_ports):
            if peering.attributes[Constants.LABELS] == stitch_port[Constants.LABELS]:
                return idx

        raise Exception(f"Did not expect peering {peering} with no stitch port")

    def handle_facility_port(self, *, sites):
        from fim.slivers.capacities_labels import Labels, Capacities

        self.sites.update(sites)

        if self.peering:
            for peering in self.peering:
                idx = self._get_stitch_info_index(peering, self.stitch_port)
                cloud = peering.attributes.get(Constants.RES_CLOUD_FACILITY)

                if not cloud:
                    cloud = self.stitch_port[idx].get(Constants.STITCH_PORT_SITE)
                    peering.attributes[Constants.RES_CLOUD_FACILITY] = cloud

            for peering in self.peering:
                idx = self._get_stitch_info_index(peering, self.stitch_port)
                cloud = peering.attributes.get(Constants.RES_CLOUD_FACILITY)
                asn = peering.attributes.get(Constants.RES_REMOTE_ASN)
                account_id = peering.attributes.get(Constants.RES_CLOUD_ACCOUNT)

                if account_id is None:
                    for discovered_stitch_info in self.discovered_stitch_infos:
                        if self.stitch_port[idx]['peer']['provider'] == discovered_stitch_info['provider']:
                            account_id = discovered_stitch_info.get("id")  # GCP's Pairing Key
                            break

                if account_id is None:
                    raise Exception(f"Missing {Constants.RES_CLOUD_ACCOUNT}:{peering}")

                subnet = peering.attributes.get(Constants.RES_LOCAL_ADDRESS)
                peer_subnet = peering.attributes.get(Constants.RES_REMOTE_ADDRESS)
                region = peering.attributes.get(Constants.RES_CLOUD_REGION)
                device = peering.attributes.get(Constants.RES_LOCAL_DEVICE)
                port = peering.attributes.get(Constants.RES_LOCAL_PORT)
                vlan = peering.attributes.get(Constants.RES_CLOUD_VLAN)
                bw = peering.attributes.get(Constants.RES_CLOUD_BW, 50)
                mtu = peering.attributes.get(Constants.RES_CLOUD_MTU)
                bgp_key = peering.attributes[Constants.RES_SECURITY]

                if not device:
                    device = self.stitch_port[idx].get(Constants.STITCH_PORT_DEVICE_NAME)

                if not port:
                    port = self.stitch_port[idx].get(Constants.STITCH_PORT_LOCAL_NAME)

                if not region:
                    region = self.stitch_port[idx].get(Constants.STITCH_PORT_REGION)

                if not cloud:
                    cloud = self.stitch_port[idx].get(Constants.STITCH_PORT_SITE)

                labels = Labels(ipv4_subnet=subnet)

                if vlan:
                    labels = Labels.update(labels, vlan=str(vlan))

                if device:
                    labels = Labels.update(labels, device_name=device)
                if port:
                    labels = Labels.update(labels, local_name=port)

                if cloud == "GCP":
                    peer_labels = Labels(ipv4_subnet=peer_subnet,
                                         asn=str(asn),
                                         bgp_key=bgp_key,
                                         account_id=account_id,
                                         # region=region,
                                         local_name='Google Cloud Platform')
                else:
                    peer_labels = Labels(ipv4_subnet=peer_subnet,
                                         asn=str(asn),
                                         bgp_key=bgp_key,
                                         account_id=account_id,
                                         region=region,
                                         local_name=cloud)

                logger.info(f"Creating_Facility_Port:Labels: {labels}")
                logger.info(f"Creating_Facility_Port:PeerLabels: {peer_labels}")

                if not mtu:
                    if cloud == "GCP":
                        mtu = 1460
                    else:
                        mtu = 9001

                capacities = Capacities(bw=int(bw), mtu=int(mtu))
                logger.info(f"Creating_Facility_Port:Capacities: {capacities}")

                facility_port = self.slice_object.add_facility_port(
                    name='Cloud-Facility-' + cloud,
                    site=cloud,
                    labels=labels,
                    peer_labels=peer_labels,
                    capacities=capacities)

                facility_port_interface = facility_port.get_interfaces()[0]
                self.facility_cloud_port_interfaces.append(facility_port_interface)
                logger.info("Done_Creating_Facility_Port:" + facility_port.toJson())
            return

        devices = set()

        for discovered_stitch_info in self.discovered_stitch_infos:
            if 'site' not in discovered_stitch_info:
                continue

            site = discovered_stitch_info[Constants.STITCH_PORT_SITE]
            assert site not in ['AWS', 'GCP']

            logger.info(
                f'{self.net_name} will use stitch info: {discovered_stitch_info}')

            device = discovered_stitch_info[Constants.STITCH_PORT_DEVICE_NAME]
            devices.add(device)
            site = discovered_stitch_info[Constants.STITCH_PORT_SITE]
            vlan = discovered_stitch_info[Constants.STITCH_PORT_VLAN]
            self.sites.add(site)
            logger.info(f"Adding Facility Port to slice: name={device}:site={site}:vlan={vlan}")

            from fabrictestbed_extensions.fablib.facility_port import FacilityPort

            fim_facility_port = self.slice_object.get_fim_topology().add_facility(
                name=device,
                site=site,
                capacities=Capacities(bw=self.bw),
                labels=Labels(vlan=str(vlan)),
            )

            facility_port = FacilityPort(self.slice_object, fim_facility_port)

            # facility_port = self.slice_object.add_facility_port(name=device,
            #                                                     site=site,
            #                                                     vlan=str(vlan))
            facility_port_interface = facility_port.get_interfaces()[0]
            self.facility_non_cloud_port_interfaces.append(facility_port_interface)

        if len(devices) != len(self.stitch_infos):
            from amscrot.policy.facility_port_handler import load_facility_info_using_stich_port
            from amscrot.policy.tag_handler import get_available_vlan

            for stitch_port in self.stitch_port:
                device = stitch_port[Constants.STITCH_PORT_DEVICE_NAME]

                if device in devices:
                    continue

                load_facility_info_using_stich_port(stitch_port)
                site = stitch_port[Constants.STITCH_PORT_SITE]
                vlan = get_available_vlan(stitch_port=stitch_port)
                self.sites.add(site)
                logger.info(f"Adding Facility Port to slice: name={device}:site={site}:vlan={vlan}")

                from fabrictestbed_extensions.fablib.facility_port import FacilityPort

                fim_facility_port = self.slice_object.get_fim_topology().add_facility(
                    name=device,
                    site=site,
                    capacities=Capacities(bw=self.bw),
                    labels=Labels(vlan=str(vlan)),
                )
                facility_port = FacilityPort(self.slice_object, fim_facility_port)

                # facility_port = self.slice_object.add_facility_port(name=device,
                #                                                     site=site,
                #                                                     vlan=str(vlan))
                facility_port_interface = facility_port.get_interfaces()[0]
                self.facility_non_cloud_port_interfaces.append(facility_port_interface)

    def handle_network(self):
        if self.facility_cloud_port_interfaces:
            tech = 'AL2S'
            port_names = [n.get_name() for n in self.facility_cloud_port_interfaces]

            logger.info(f"Creating L3VPN Network:{self.net_name}:FacilityPorts:{port_names}:techonolgy={tech}")

            self.net = self.slice_object.add_l3network(name=self.net_name,
                                                       interfaces=self.facility_cloud_port_interfaces,
                                                       type='L3VPN',
                                                       technology=tech)

        if self.facility_non_cloud_port_interfaces:
            if len(self.sites) == 2:
                net_type = 'L2STS'
            elif len(self.sites) == 1:
                net_type = 'L2Bridge'
            else:
                raise Exception(f"The number of sites in {self.sites} is not supported")

            port_names = [i.get_name() for i in self.facility_non_cloud_port_interfaces]
            logger.info(f"Creating {net_type} Network:{self.net_name}:FacilityPorts:{port_names}")
            # self.net = self.slice_object.add_l2network(name=self.net_name, type=net_type)
            self.net = self.slice_object.add_l2network(name=self.net_name)

            # We loop for now. Maybe we can remove in a later release post 1.7.3.
            for itf in self.facility_non_cloud_port_interfaces:
                self.net.add_interface(itf)

    def build(self) -> FabricNetwork:
        assert self.net
        return FabricNetwork(label=self.label, delegate=self.net, layer3=self.layer3,
                             peering=self.peering, peer_layer3=self.peer_layer3)



import logging
from argparse import ArgumentParser, RawDescriptionHelpFormatter

from amscrot.util.constants import Constants


def create_parser(usage='%(prog)s [options]',
                  description='',
                  formatter_class=None):
    formatter_class = formatter_class or RawDescriptionHelpFormatter
    return ArgumentParser(usage=usage, description=description, formatter_class=formatter_class)


def build_parser(*, manage_workflow, manage_sessions, display_stitch_info):
    description = (
        'AmSCROT'
        '\n'
        '\n'
        'Examples:'
        '\n'
        "      amscrot workflow --var-file vars.yml --session test-chi -validate"
        '\n'
        "      amscrot workflow --config-dir . --session test-chi -validate"
        '\n'
        '      amscrot stitch-policy -providers "fabric,sense"'
    )

    parser = create_parser(description=description)
    subparsers = parser.add_subparsers()
    workflow_parser = subparsers.add_parser('workflow', help='Manage AmSCROT workflows')
    workflow_parser.add_argument('-c', '--config-dir', type=str, default='.',
                                 help='config directory with .fab files. Defaults to current directory.',
                                 required=False)
    workflow_parser.add_argument('-v', '--var-file', type=str, default='',
                                 help="Yaml file with key-value pairs to override the variables' default values",
                                 required=False)
    workflow_parser.add_argument('-s', '--session', type=str, default='',
                                 help='friendly session name to help track a workflow', required=True)
    workflow_parser.add_argument('-p', '--policy-file', type=str, default='',
                                 help="Yaml stitching policy file",
                                 required=False)
    workflow_parser.add_argument('-validate', action='store_true', default=False,
                                 help='assembles and validates all .fab files  in the config directory')
    workflow_parser.add_argument(
        '-apply', action='store_true', default=False,
        help='create resources. The -apply is repeatable and is designed to converge to the desired state')

    workflow_parser.add_argument('-init', action='store_true', default=False, help='display resource ordering')
    workflow_parser.add_argument('-stitch-info', action='store_true', default=False, help='display network stitch-info')
    workflow_parser.add_argument('-plan', action='store_true', default=False, help='shows plan')
    workflow_parser.add_argument('-use-remote-policy', action='store_true', default=False, help='use remote policy')
    workflow_parser.add_argument('-show', action='store_true', default=False, help='display resource.')
    workflow_parser.add_argument('-summary', action='store_true', default=False,
                                 help='display summary. used with -show')
    workflow_parser.add_argument('-stats', action='store_true', default=False, help='display stats')
    workflow_parser.add_argument('-json', action='store_true', default=False,
                                 help='use json output. relevant when used with -show or -plan')
    workflow_parser.add_argument(
        '-destroy', action='store_true', default=False,
        help='delete resources. The -destroy is repeatable and is designed to converge to the desired state')

    workflow_parser.set_defaults(dispatch_func=manage_workflow)

    sessions_parser = subparsers.add_parser('sessions', help='Manage AmSCROT sessions ')
    sessions_parser.add_argument('-show', action='store_true', default=False, help='display sessions')
    sessions_parser.add_argument('-json', action='store_true', default=False, help='use json format')
    sessions_parser.set_defaults(dispatch_func=manage_sessions)
    stitch_parser = subparsers.add_parser('stitch-policy', help='Display stitch policy between two poviders')
    stitch_parser.add_argument('-providers', type=str, required=True,
                               default="", help='two comma separated providers from chi,fabric,cloudlab, or sense')
    stitch_parser.add_argument('-c', '--credential-file', type=str, default='~/.amscrot/amscrot_credentials.yml',
                               help='AmSCROT credential file. Defaults to ~/.amscrot/amscrot_credentials.yml',
                               required=False)
    stitch_parser.add_argument('-p', '--policy-file', type=str, default='',
                               help="Yaml stitching policy file",
                               required=False)
    stitch_parser.add_argument('--profile', type=str, default='fabric',
                               help="fabric profile from credential file. Defaults to fabric",
                               required=False)
    stitch_parser.add_argument('-use-remote-policy', action='store_true', default=False, help='use remote policy')
    stitch_parser.set_defaults(dispatch_func=display_stitch_info)
    return parser


def get_log_level():
    import os
    return os.environ.get('AMSCROT_LOG_LEVEL', "INFO")


def get_log_location():
    import os
    return os.environ.get('AMSCROT_LOG_LOCATION', "/tmp/amscrot.log")


def get_formatter():
    fmt = "%(asctime)s [%(filename)s:%(lineno)d] [%(levelname)s] %(message)s"
    return logging.Formatter(fmt)


def disable_stdout_logs():
    import os
    return os.environ.get('AMSCROT_DISABLE_LOGS_TO_STDOUT', "false").lower() == 'true'


_LOGGER = None


def get_logger():
    global _LOGGER

    if not _LOGGER:
        _LOGGER = init_logger()

    return _LOGGER


def init_logger():
    from logging.handlers import RotatingFileHandler

    log_config = {'log-file': get_log_location(),
                  'log-level': get_log_level(),
                  'log-retain': 5,
                  'log-size': 5000000,
                  'logger': 'amscrot'}

    logger = logging.getLogger(str(log_config.get(Constants.PROPERTY_CONF_LOGGER, __name__)))
    log_level = log_config.get(Constants.PROPERTY_CONF_LOG_LEVEL, "INFO")
    logger.setLevel(log_level)

    formatter = get_formatter()
    file_handler = RotatingFileHandler(log_config.get(Constants.PROPERTY_CONF_LOG_FILE),
                                       backupCount=int(log_config.get(Constants.PROPERTY_CONF_LOG_RETAIN)),
                                       maxBytes=int(log_config.get(Constants.PROPERTY_CONF_LOG_SIZE)))

    file_handler.setFormatter(formatter)
    logger.propagate = False
    logger.addHandler(file_handler)

    if not disable_stdout_logs():
        stream_handler = logging.StreamHandler()
        stream_handler.setFormatter(formatter)
        logger.addHandler(stream_handler)

    global _LOGGER

    _LOGGER = logger
    return logger


def absolute_path(path):
    from pathlib import Path
    import os

    path = Path(path).expanduser().absolute()

    return os.path.realpath(str(path))


def load_as_ns_from_yaml(*, dir_path=None, content=None):
    import yaml
    import json
    from types import SimpleNamespace

    objs = []

    if dir_path:
        from pathlib import Path
        import os

        dir_path = Path(dir_path).expanduser().absolute()

        if not os.path.isdir(dir_path):
            raise Exception(f'Expected a directory {dir_path}')

        configs = [conf for conf in os.listdir(dir_path) if conf.endswith(Constants.FAB_EXTENSION)]

        if not configs:
            raise Exception(f'No {Constants.FAB_EXTENSION} config files found in  {dir_path}')

        for config in configs:
            file_name = os.path.join(dir_path, config)

            with open(file_name, 'r') as stream:
                loader = yaml.FullLoader
                obj = yaml.load(stream, Loader=loader)
                obj = json.loads(json.dumps(obj), object_hook=lambda dct: SimpleNamespace(**dct))
                objs.append(obj)
    else:
        obj = yaml.safe_load(content)
        obj = json.loads(json.dumps(obj), object_hook=lambda dct: SimpleNamespace(**dct))
        objs.append(obj)

    return objs


def load_yaml_from_file(file_name):
    import yaml
    from pathlib import Path

    path = Path(file_name).expanduser().absolute()

    with open(str(path), 'r') as stream:
        return yaml.load(stream, Loader=yaml.FullLoader)


def load_vars(var_file):
    import yaml
    import os

    if not os.path.isfile(var_file):
        raise Exception(f'The supplied var-file {var_file} is invalid')

    with open(var_file, 'r') as stream:
        return yaml.load(stream, Loader=yaml.FullLoader)


def get_stats_base_dir(friendly_name):
    from pathlib import Path
    import os

    base_dir = os.path.join(str(Path.home()), '.amscrot', 'stats', friendly_name)
    os.makedirs(base_dir, exist_ok=True)
    return base_dir


def get_base_dir(friendly_name):
    from pathlib import Path
    import os

    base_dir = os.path.join(str(Path.home()), '.amscrot', 'sessions', friendly_name)
    os.makedirs(base_dir, exist_ok=True)
    return base_dir


def get_inventory_dir(friendly_name):
    import os
    inv_dir = os.path.join(get_base_dir(friendly_name), "inventory")
    os.makedirs(inv_dir, exist_ok=True)
    return inv_dir


def get_ssh_dir(friendly_name):
    import os
    ssh_dir = os.path.join(get_base_dir(friendly_name), "ssh")
    os.makedirs(ssh_dir, exist_ok=True)
    return ssh_dir


def dump_sessions(to_json: bool):
    from amscrot.util import state as sutil
    from pathlib import Path
    import os
    import sys

    base_dir = os.path.join(str(Path.home()), '.amscrot', 'sessions')
    os.makedirs(base_dir, exist_ok=True)
    sessions = os.listdir(base_dir)
    sessions = [dict(session=s, config_dir=sutil.load_meta_data(s, 'config_dir')) for s in sessions]

    if to_json:
        import json

        sys.stdout.write(json.dumps(sessions, default=lambda o: o.__dict__, indent=3))
    else:
        import yaml

        sys.stdout.write(yaml.dump(sessions))

    return sessions


def get_counters(*, states):
    nodes = networks = services = pending = failed = 0

    for state in states:
        pending += len(state.pending)
        pending += len(state.pending_internal)
        nodes += len([n for n in state.node_states])
        networks += len([n for n in state.network_states])
        services += len(state.service_states)

        for key, detail in state.creation_details.items():
            failed += detail['failed_count']

    return nodes, networks, services, pending, failed


def can_read(path: str):
    from pathlib import Path

    path = str(Path(path).expanduser().absolute())

    try:
        with open(path, 'r') as f:
            f.read()
        return True
    except:
        return False


def can_read_json(path: str):
    try:
        with open(path, 'r') as fp:
            import json

            json.load(fp)

        return True
    except:
        return False


# noinspection PyBroadException
def is_private_key(private_key_file: str):
    import paramiko

    try:
        paramiko.RSAKey.from_private_key_file(private_key_file)
        return True
    except Exception:
        pass

    try:
        paramiko.ecdsakey.ECDSAKey.from_private_key_file(private_key_file)
        return True
    except Exception:
        pass

    return False


def generate_bgp_key_if_needed(friendly_name):
    from amscrot.model.state import get_loader, get_dumper
    import yaml
    import os

    file_path = os.path.join(get_base_dir(friendly_name), friendly_name + '_secrets.yml')
    secrets = dict()

    if os.path.exists(file_path):
        with open(file_path, 'r') as stream:
            try:
                secrets = yaml.load(stream, Loader=get_loader())
            except Exception as e:
                from amscrot.exceptions import StateException

                raise StateException(f'Exception loading secrets at {file_path}:{e}')

    if Constants.RES_BGP_KEY in secrets:
        return secrets[Constants.RES_BGP_KEY]

    import uuid
    import hashlib

    md = hashlib.md5()
    temp = str(uuid.uuid4())
    md.update(temp.encode('utf-8'))
    secrets[Constants.RES_BGP_KEY] = md.hexdigest()

    with open(file_path, "w") as stream:
        try:
            stream.write(yaml.dump(secrets, Dumper=get_dumper()))
        except Exception as e:
            from amscrot.exceptions import StateException

            raise StateException(f'Exception saving secrets in file {file_path}:{e}')

    return secrets[Constants.RES_BGP_KEY]


def normalize_alias(alias):
    """

    :param alias:
    :return:
    """
    undesired = "'\",!~+*^%@#$();/\\ "
    translation_table = str.maketrans(undesired, ":" * len(undesired))
    alias = alias.translate(translation_table)
    alias = alias.replace(":", "")
    alias = alias.replace("_", "-")
    return alias.lower()



from amscrot.util.constants import Constants


def has_resolved_external_dependencies(*, resource, attribute):
    resolved_dependencies = [rd for rd in resource[Constants.RESOLVED_EXTERNAL_DEPENDENCIES] if rd.attr == attribute]
    return len(resolved_dependencies) > 0


def has_resolved_internal_dependencies(*, resource, attribute):
    resolved_dependencies = [rd for rd in resource[Constants.RESOLVED_INTERNAL_DEPENDENCIES] if rd.attr == attribute]
    return len(resolved_dependencies) > 0


def get_values_for_dependency(*, resource, attribute):
    assert has_resolved_external_dependencies(resource=resource, attribute=attribute) \
           or has_resolved_internal_dependencies(resource=resource, attribute=attribute)

    values = []

    for temp in resource.get(attribute):
        for value in temp:
            values.append(value)

    return values

def get_single_value_for_dependency_if_any(*, resource, attribute):
    values = get_values_for_dependency(resource=resource, attribute=attribute)

    return values[0]

def get_single_value_for_dependency(*, resource, attribute):
    values = get_values_for_dependency(resource=resource, attribute=attribute)
    assert len(values) == 1
    return values[0]



from types import SimpleNamespace
from typing import List

from amscrot.util.config_models import *
from amscrot.util.config_models import ResourceConfig


class ResourceDependencyEvaluator:
    def __init__(self, resources: List[ResourceConfig], providers: List[ProviderConfig]):
        self.resources = resources
        self.providers = providers
        self.dependency_map: Dict[ResourceConfig, Set[ResourceConfig]] = {}

    def _find_resource_for(self, basic_config, res):
        temp = resource_from_basic_config(basic_config, self.providers)
        if temp not in self.resources:
            from amscrot.exceptions import ParseConfigException

            raise ParseConfigException(
                f'{temp.label} not found. {res.label} depends on it. Maybe its count is set to zero?')

        index = self.resources.index(resource_from_basic_config(basic_config, self.providers))
        return self.resources[index]

    def add_dependency(self, res: ResourceConfig, key: str, dependency_info: DependencyInfo):
        found = self._find_resource_for(dependency_info.resource, res)
        is_external = res.provider.label != found.provider.label
        temp = Dependency(key=key, resource=found, attribute=dependency_info.attribute, is_external=is_external)
        res.add_dependency(temp)
        self.dependency_map[res].add(found)

    def handle_dependency(self, resource: ResourceConfig, key: str, value):
        if isinstance(value, DependencyInfo):
            self.add_dependency(resource, key, value)
        elif isinstance(value, list):
            for val in value:
                self.handle_dependency(resource, key, val)
        elif isinstance(value, dict):
            prefix = key + '.'
            for k, val in value.items():
                self.handle_dependency(resource, prefix + k, val)
        elif isinstance(value, SimpleNamespace):
            raise Exception(key)

    def evaluate(self) -> Dict[ResourceConfig, Set[ResourceConfig]]:
        for resource in self.resources:
            self.dependency_map[resource] = set()

            for key, value in resource.attributes.items():
                self.handle_dependency(resource, key, value)

        return self.dependency_map


def order_resources(dependency_map:  Dict[ResourceConfig, Set[ResourceConfig]]) -> List[ResourceConfig]:
    ordered_resources: List[ResourceConfig] = []
    while len(dependency_map) > 0:
        found = None

        for key, value in dependency_map.items():
            if len(value) == 0:
                found = key
                break

        if not found:
            raise Exception("circular dependencies ...." + str(dependency_map))

        dependency_map.pop(found)
        ordered_resources.append(found)
        for value in dependency_map.values():
            if found in value:
                value.remove(found)

    return ordered_resources



from amscrot.exceptions import ResourceTypeNotSupported
from amscrot.provider.api.provider import Provider
from amscrot.util.constants import Constants
from amscrot.util.utils import get_logger
from . import aws_constants

logger = get_logger()


class AwsProvider(Provider):
    def __init__(self, *, type, label, name, config: dict):
        super().__init__(type=type, label=label, name=name, logger=logger, config=config)
        self.supported_resources = [Constants.RES_TYPE_NETWORK.lower()]

    @property
    def access_key(self):
        return self.config.get(aws_constants.ACCESS_KEY)

    @property
    def secret_key(self):
        return self.config.get(aws_constants.SECRET_KEY)

    def setup_environment(self):
        from amscrot.util import utils

        credential_file = self.config.get(Constants.CREDENTIAL_FILE)
        profile = self.config.get(Constants.PROFILE)
        config = utils.load_yaml_from_file(credential_file)

        if profile not in config:
            from amscrot.exceptions import ProviderException

            raise ProviderException(
                f"credential file {credential_file} does not have a section for keyword {profile}"
            )

        self.config = config[profile]
        normalized_config = {}

        for k, v in self.config.items():
            normalized_config[k.upper()] = v

        self.config = normalized_config
        assert self.access_key
        assert self.secret_key

    def _handle_peering_config(self, resource):
        import amscrot.provider.api.dependency_util as util
        from amscrot.model import Network

        peering = resource.get(Constants.RES_PEERING)
        prop = 'stitch_interface'
        assert Constants.RES_SECURITY in peering.attributes

        if peering and util.has_resolved_external_dependencies(resource=resource, attribute=prop):
            net = util.get_single_value_for_dependency(resource=resource, attribute=prop)

            if isinstance(net, Network):
                iface = net.interface

                if isinstance(iface, list):
                    iface = iface[0]

                if isinstance(iface, dict) and iface.get("provider") == "fabric":
                    peering.attributes[Constants.RES_ID] = iface[Constants.RES_ID]
                    self.logger.info(f"Added id to peering config:id={iface[Constants.RES_ID]}")

        return peering

    def do_add_resource(self, *, resource: dict):
        label = resource.get(Constants.LABEL)
        rtype = resource.get(Constants.RES_TYPE)

        if rtype not in self.supported_resources:
            raise ResourceTypeNotSupported(f"{rtype} for {label}")

        name_prefix = resource.get(Constants.RES_NAME_PREFIX)
        net_name = f'{self.name}-{name_prefix}'
        layer3 = resource.get(Constants.RES_LAYER3)
        peering = self._handle_peering_config(resource)

        from .aws_network import AwsNetwork
        from amscrot.policy.policy_helper import get_stitch_port_for_provider

        stitch_port = get_stitch_port_for_provider(resource=resource, provider=self.type)
        net = AwsNetwork(label=label, name=net_name, provider=self,
                         layer3=layer3, peering=peering, stitch_port=stitch_port)
        self._networks.append(net)

        if self.resource_listener:
            self.resource_listener.on_added(source=self, provider=self, resource=net)

    def do_create_resource(self, *, resource: dict):
        rtype = resource.get(Constants.RES_TYPE)
        assert rtype in self.supported_resources
        label = resource.get(Constants.LABEL)

        for net in [net for net in self._networks if net.label == label]:
            self.logger.debug(f"Creating network: {vars(net)}")
            net.create()

            if self.resource_listener:
                self.resource_listener.on_created(source=self, provider=self, resource=net)

            self.logger.debug(f"Created network: {vars(net)}")

    def do_delete_resource(self, *, resource: dict):
        rtype = resource.get(Constants.RES_TYPE)
        assert rtype in self.supported_resources
        label = resource.get(Constants.LABEL)

        if rtype == Constants.RES_TYPE_NODE.lower():
            # DO NOTHING
            return

        states = resource.get(Constants.SAVED_STATES, [])
        state = next(filter(lambda s: s.label == label, states), None)
        net_name = f'{self.name}-{resource.get(Constants.RES_NAME_PREFIX)}'
        logger.debug(f"Deleting network: {net_name}")

        from .aws_network import AwsNetwork
        from amscrot.policy.policy_helper import get_stitch_port_for_provider

        stitch_port = get_stitch_port_for_provider(resource=resource, provider=self.type)
        layer3 = resource.get(Constants.RES_LAYER3)
        peering = resource.get(Constants.RES_PEERING)
        net = AwsNetwork(label=label, name=net_name, provider=self, layer3=layer3, peering=peering, stitch_port=stitch_port, state=state)
        net.delete()
        logger.debug(f"Done Deleting network: {net_name}")

        if self.resource_listener:
            self.resource_listener.on_deleted(source=self, provider=self, resource=net)



from fabrictestbed_extensions.fablib.node import Node as Delegate
from fabrictestbed_extensions.fablib.slice import Slice

from amscrot.model import Node
from amscrot.util.constants import Constants
from amscrot.util.utils import get_logger
from .fabric_constants import *

logger = get_logger()


class FabricNode(Node):
    def __init__(self, *, label, delegate: Delegate, nic_model: str, network_label: str):
        flavor = {'cores': delegate.get_cores(), 'ram': delegate.get_ram(), 'disk': delegate.get_disk()}
        super().__init__(label=label, name=delegate.get_name(), image=delegate.get_image(), site=delegate.get_site(),
                         flavor=str(flavor))
        logger.info(f" Node {self.name} constructor called ... ")
        self._delegate = delegate
        self.nic_model = nic_model
        self._slice_object = delegate.get_slice()
        self.slice_name = self._slice_object.get_name()
        self.mgmt_ip = delegate.get_management_ip()
        self.mgmt_ip = str(self.mgmt_ip) if self.mgmt_ip else None
        self.network_label = network_label
        self.username = delegate.get_username()
        self.user = self.username
        self.state = delegate.get_reservation_state()
        self.state = self.state.lower() if self.state else None
        self.host = self.mgmt_ip
        self.keyfile = self._delegate.get_private_key_file()
        self.jump_user = self._delegate.get_fablib_manager().get_bastion_username()
        self.jump_host = self._delegate.get_fablib_manager().get_bastion_host()
        self.jump_keyfile = self._delegate.get_fablib_manager().get_bastion_key_location()
        self._used_dataplane_ipv4 = None
        self.dataplane_ipv4 = None
        self.dataplane_ipv6 = None
        self.id = delegate.get_reservation_id()
        self.components = [dict(name=c.get_name(), model=c.get_model()) for c in delegate.get_components()]
        self.addr_list = {}

    def handle_networking(self):
        if not self.mgmt_ip:
            logger.warning(f" Node {self.name} has no management ip ")
            return

        itfs = [itf for itf in self.delegate.get_interfaces() if itf.get_ip_addr()]

        for itf in itfs:
            if FABRIC_STITCH_NET_IFACE_NAME in itf.get_name():
                self.dataplane_ipv4 = str(itf.get_ip_addr())
                break
            elif f'FABNET_IPv4_{self.site}' in itf.get_name():
                self.dataplane_ipv4 = str(itf.get_ip_addr())
                break

        for itf in itfs:
            if f'FABNET_IPv6_{self.site}' in itf.get_name():
                node_addr = itf.get_ip_addr()
                self.dataplane_ipv6 = str(node_addr)
                break
            elif f'FABNET_IPv6Ext_{self.site}' in itf.get_name():
                node_addr = itf.get_ip_addr()
                self.dataplane_ipv6 = str(node_addr)
                break

    @property   
    def delegate(self) -> Delegate:
        return self._delegate

    def set_network_label(self, network_label):
        self.network_label = network_label

    def used_dataplane_ipv4(self):
        return self._used_dataplane_ipv4

    def set_used_dataplane_ipv4(self, used_dataplane_ipv4):
        self._used_dataplane_ipv4 = used_dataplane_ipv4

    def get_interfaces(self):
        return self._delegate.get_interfaces()

    def get_interface(self, *, network_name):
        assert network_name
        return self._delegate.get_interface(network_name=network_name)

    def upload_file(self, local_file_path, remote_file_path, retry=3, retry_interval=10):
        self._delegate.upload_file(local_file_path, remote_file_path, retry, retry_interval)

    def download_file(self, local_file_path, remote_file_path, retry=3, retry_interval=10):
        self._delegate.download_file(local_file_path, remote_file_path, retry, retry_interval)

    def upload_directory(self, local_directory_path, remote_directory_path, retry=3, retry_interval=10):
        self._delegate.upload_directory(local_directory_path, remote_directory_path, retry, retry_interval)

    def download_directory(self, local_directory_path, remote_directory_path, retry=3, retry_interval=10):
        self._delegate.download_directory(local_directory_path, remote_directory_path, retry, retry_interval)

    def execute(self, command, retry=3, retry_interval=10):
        self._delegate.execute(command, retry, retry_interval)

    def add_route(self, subnet, gateway):
        self._delegate.ip_route_add(subnet=subnet, gateway=gateway)

    def get_management_ip(self) -> str:
        return self._delegate.get_management_ip()

    def get_dataplane_address(self, network=None, interface=None, af=Constants.IPv4):
        if af == Constants.IPv4:
            return self.dataplane_ipv4
        elif af == Constants.IPv6:
            return self.dataplane_ipv6
        else:
            return None

    def get_reservation_id(self) -> str:
        return self._delegate.get_reservation_id()

    def get_reservation_state(self) -> str:
        return self._delegate.get_reservation_state()


class NodeBuilder:
    def __init__(self, label, slice_object: Slice, name: str,  resource: dict):
        from .fabric_constants import FABRIC_RANDOM
        site = resource.get(Constants.RES_SITE, FABRIC_RANDOM)

        if site == FABRIC_RANDOM:
            from fabrictestbed_extensions.fablib.fablib import fablib

            site = fablib.get_random_site()

        image = resource.get(Constants.RES_IMAGE, Delegate.default_image)
        flavor = resource.get(Constants.RES_FLAVOR, {'cores': 2, 'ram': 8, 'disk': 10})
        cores = flavor.get(Constants.RES_FLAVOR_CORES, Delegate.default_cores)
        ram = flavor.get(Constants.RES_FLAVOR_RAM, Delegate.default_ram)
        disk = flavor.get(Constants.RES_FLAVOR_DISK, Delegate.default_disk)
        self.label = label
        self.nic_model = resource.get(Constants.RES_NIC_MODEL, 'NIC_Basic')
        self.node: Delegate = slice_object.add_node(name=name, image=image, site=site, cores=cores, ram=ram, disk=disk)

        # Use fully automated ip v4
        if resource.get("enable_fabnetv4", True) and resource.get(Constants.RES_TYPE_NETWORK) is None:
            self.node.add_fabnet(net_type="IPv4", nic_type=self.nic_model)

        # Use fully automated ip v6
        if resource.get("enable_fabnetv6", False):
            self.node.add_fabnet(net_type="IPv6", nic_type=self.nic_model)

        # Use fully automated ip v6Ext
        if resource.get("enable_fabnetv6ext", False):
            self.node.add_fabnet(net_type="IPv6Ext", nic_type=self.nic_model,
                                 routes=resource.get("routes", None))

    def add_component(self, model=None, name=None):
        self.node.add_component(model=model, name=name)

    def build(self) -> FabricNode:
        return FabricNode(label=self.label, delegate=self.node, nic_model=self.nic_model, network_label="")



from amscrot.model import Network
from amscrot.util.utils import get_logger, Constants
from . import aws_utils
from .aws_provider import AwsProvider
from .aws_exceptions import AwsException

logger = get_logger()


class AwsNetwork(Network):
    def __init__(self, *, label, name: str, provider: AwsProvider, layer3, peering, stitch_port, state=None):
        super().__init__(label=label, name=name, site="")
        self._provider = provider
        self.layer3 = layer3
        self.peering = peering
        self.stitch_port = stitch_port
        self.vpc_id = None
        self.vpn_id = None
        self.vif_details = {}
        self.route_table_details = {}
        self._state = state

    @property
    def vpn_gateway_name(self):
        return f'{self._provider.name}-fab-vpn-gateway'

    @property
    def vif_name(self):
        return f'{self._provider.name}-fab-vif'

    @property
    def connection_name(self):
        return self.peering.attributes.get(Constants.RES_ID)

    def create(self):
        region = self.peering.attributes.get(Constants.RES_CLOUD_REGION)

        if not region:
           region = self.stitch_port['peer'].get(Constants.STITCH_PORT_REGION)

        if not region:
           raise AwsException(f"Missing cloud region")

        logger.info(f'{self.name} using region {region}')

        ec2_client = aws_utils.create_ec2_client(region=region,
                                                 access_key=self._provider.access_key,
                                                 secret_key=self._provider.secret_key)
        self.vpc_id = self.peering.attributes.get(Constants.RES_CLOUD_VPC)

        if not self.vpc_id:
            raise AwsException(f"must supply vpc id using peering attribute: {Constants.RES_CLOUD_VPC}")

        if not aws_utils.is_vpc_available(ec2_client=ec2_client, vpc_id=self.vpc_id):
            raise AwsException(f"Vpc is not available:{self.vpc_id}")

        logger.info(f'Vpc {self.vpc_id} is available')

        direct_connect_client = aws_utils.create_direct_connect_client(
            region=region,
            access_key=self._provider.access_key,
            secret_key=self._provider.secret_key)

        connection_id, vlan = aws_utils.find_available_dx_connection(
            direct_connect_client=direct_connect_client,
            name=self.connection_name)

        logger.info(f'connection {self.connection_name} is available:connection_id={connection_id}')

        amazon_asn = self.peering.attributes.get(Constants.RES_REMOTE_ASN)

        if isinstance(amazon_asn, str):
           amazon_asn = int(amazon_asn)

        self.vpn_id = aws_utils.find_vpn_gateway(ec2_client=ec2_client, name=self.vpn_gateway_name)

        if not self.vpn_id:
            logger.info(f'Creating vpn gateway: name={self.vpn_gateway_name}:vpc={self.vpc_id}')
            self.vpn_id = aws_utils.create_vpn_gateway(
                ec2_client=ec2_client,
                name=self.vpn_gateway_name,
                amazon_asn=amazon_asn)
            logger.info(f'Created vpn gateway: name={self.vpn_gateway_name}')
        else:
            logger.info(f'Found vpn gateway: name={self.vpn_gateway_name}')

        aws_utils.attach_vpn_gateway_if_needed(ec2_client=ec2_client, vpn_id=self.vpn_id, vpc_id=self.vpc_id)

        self.route_table_details = aws_utils.create_route_table_if_needed(ec2_client=ec2_client,
                                                                    cidr=self.layer3.attributes['subnet'],
                                                                    vpc_id=self.vpc_id,
                                                                    vpn_id=self.vpn_id)

        self.vif_details = aws_utils.create_private_virtual_interface(
            direct_connect_client=direct_connect_client,
            vpn_gateway_id=self.vpn_id,
            connection_id=connection_id,
            vlan=vlan,
            peering=self.peering,
            vif_name=f"{self.vif_name}")

    def delete(self):
        region = self.peering.attributes.get(Constants.RES_CLOUD_REGION)

        if not region:
           region = self.stitch_port['peer'].get(Constants.STITCH_PORT_REGION)

        if not region:
           raise AwsException(f"Missing cloud region")

        ec2_client = aws_utils.create_ec2_client(region=region,
                                                 access_key=self._provider.access_key,
                                                 secret_key=self._provider.secret_key)

        vpc_id = self.peering.attributes.get(Constants.RES_CLOUD_VPC)
        if not vpc_id:
            return

        if not aws_utils.is_vpc_available(ec2_client=ec2_client, vpc_id=vpc_id):
            return

        vpn_id = aws_utils.find_vpn_gateway(ec2_client=ec2_client, name=self.vpn_gateway_name)

        if not vpn_id:
            return

        route_table_details = {}

        if self._state:
            route_table_details = self._state.attributes.get('route_table_details', {})

        route_table_id = route_table_details.get('RouteTableId')
        aws_utils.delete_route_table_if_needed(ec2_client=ec2_client, vpc_id=vpc_id, route_table_id=route_table_id)

        direct_connect_client = aws_utils.create_direct_connect_client(
            region=region,
            access_key=self._provider.access_key,
            secret_key=self._provider.secret_key)

        aws_utils.delete_private_virtual_interface(direct_connect_client=direct_connect_client,
                                                   vif_name=f"{self.vif_name}")

        aws_utils.detach_vpn_gateway_if_needed(ec2_client=ec2_client, vpn_id=vpn_id, vpc_id=vpc_id)
        aws_utils.delete_vpn_gateway(ec2_client=ec2_client, name=self.vpn_gateway_name)



import logging

from amscrot.model import Service, Node, SSHNode, Network
from amscrot.provider.api.provider import Provider
from amscrot.util.constants import Constants

'''

To add a provider, all you should need is to add its classpath to amscrot.util.constants.Constants.PROVIDER_CLASSES


see tests/examples/dummy-service for a simple example.
see tests/examples/dummy-service-dependency for en example with an external dependency. (Dependency across providers)

This example also show how to hide attributes by using underscores and how to expose interesting attributes
to other resources by extracting it and using yaml python supported types.

Useful Commands:
see tests/examples/dummy-service  # for a simple example
see tests/examples/dummy-service/config.fab 

>amscrot workflow --session <session> -validate
>amscrot workflow --session <session> -plan
>amscrot workflow --session <session> -apply
>amscrot workflow --session <session> -show
>amscrot workflow --session <session> -destroy
'''


class HideAttribute:
    def __init__(self, *, x=10, y=20):
        self.x = x
        self.y = y

class DummyNode(Node, SSHNode):
    def __init__(self, *, label, name: str, image, site, flavor, logger: logging.Logger):
        super().__init__(label=label, name=name, image=image, site=site, flavor=flavor)
        self.logger = logger

    def create(self):
        self.logger.info(f" Dummy Node {self.name} created.")
        self.user = "dummy_user"
        self.host = f"localhost_{self.name}"
        self.keyfile = "~/.ssh/dummy_key"
        #self.jump_user = "dummy_jump_user"
        self.jump_host = "dummy.jumphost.net"
        self.jump_keyfile = "~/.ssh/dummy_jump_key"

    def delete(self):
        self.logger.info(f" Dummy Node {self.name} deleted")

    def get_reservation_id(self):
        pass

    def get_reservation_state(self):
        pass

    def add_route(self, subnet, gateway):
        pass

    def get_dataplane_address(self, network=None, interface=None, af=None):
        return "192.168.1.10"

class DummyNetwork(Network):
    def __init__(self, *, label, name: str, site, logger: logging.Logger):
        super().__init__(label=label, name=name, site=site)
        self.logger = logger
        
    def create(self):
        self.logger.info(f" Dummy Network {self.name} created at site {self.site}")
        
    def delete(self):
        self.logger.info(f" Dummy Network {self.name} deleted")
        
    def get_reservation_id(self):
        return f"res-{self.name}"

class DummyService(Service):
    def __init__(self, *, label, name: str, image, x=None, logger: logging.Logger):
        super().__init__(label=label, name=name)
        self.logger = logger
        self.image = image
        # We hide a complex attribute by using underscores otherwise marshalling the state will fail
        # as we do not have yaml representer for it.
        self._hidden_attribute = HideAttribute(x=x)
        self.exposed_attribute_x = self._hidden_attribute.x  # Here we expose x which can be marshalled

    def create(self):
        import random

        if not self.exposed_attribute_x:
            self._hidden_attribute = HideAttribute(x=random.randint(1, 5), y=random.random())
            self.exposed_attribute_x = self._hidden_attribute.x

        self.logger.info(f" Dummy Service {self.name} created. X={self.exposed_attribute_x}")

    def delete(self):
        self.logger.info(f" Dummy Service {self.name} deleted")

from amscrot.util.utils import get_logger

logger: logging.Logger = get_logger()


class DummyProvider(Provider):

    def setup_environment(self):
        pass

    def __init__(self, *, type, label, name, config: dict):
        super().__init__(type=type, label=label, name=name, logger=logger, config=config)

    def _validate_resource(self, resource: dict):
        assert resource.get(Constants.LABEL)
        assert resource.get(Constants.RES_TYPE) in Constants.RES_SUPPORTED_TYPES
        assert resource.get(Constants.RES_NAME_PREFIX)
        assert resource.get(Constants.RES_COUNT, 1)

        label = resource.get(Constants.LABEL)
        self.logger.info(f"Validated:OK Resource={label} using {self.label}")

    def do_add_resource(self, *, resource: dict):
        """
        Called by add_resource(self, *, resource: dict) if resource has no external dependencies.
        The add_resource(self, *, resource: dict) puts resources in the pending dictionary when
        they have external dependencies.

        When the external dependencies are satisfied following a resource creation event, this method
        would be called automatically. See on_created(self, *, source, provider, resource: object)

        Note that external dependencies are respurce dependencies across different providers.
        @param resource: resource attributes
        """
        label = resource.get(Constants.LABEL)
        self.logger.info(f"Adding resource={label} using {self.label}")
        self._validate_resource(resource)
        image = resource.get(Constants.RES_IMAGE)
        exposed_attribute_x = resource.get("exposed_attribute_x")

        # In the dependency example, exposed_attribute_x is an external dependency
        import amscrot.provider.api.dependency_util as util

        if util.has_resolved_external_dependencies(resource=resource, attribute='exposed_attribute_x'):
            # Service dtn1 depends on service dtn2. dtn1 and dtn2 have different providers.
            # Extract all values. There can be more than one if count of service.dtn2 is greater than 1.

            values = util.get_values_for_dependency(resource=resource, attribute='exposed_attribute_x')
            assert values

            # The next line handles both dependencies:
            #          exposed_attribute_x: "{{ service.dtn2 }}"
            #          exposed_attribute_x: "{{ service.dtn2.exposed_attribute_x }}"
            values = [value.exposed_attribute_x if isinstance(value, DummyService) else value for value in values]
            exposed_attribute_x = sum(values)

        label = resource.get(Constants.LABEL)
        rtype = resource.get(Constants.RES_TYPE)

        if rtype == Constants.RES_TYPE_NODE.lower():
            node_count = resource.get(Constants.RES_COUNT, 1)
            name_prefix = resource.get(Constants.RES_NAME_PREFIX)
            site = resource.get(Constants.RES_SITE)
            flavor = resource.get(Constants.RES_FLAVOR, "dummy_flavor")

            for i in range(node_count):
                name = f"{name_prefix}{i}"
                node = DummyNode(label=label, name=name, site=site, image=image, flavor=flavor, logger=self.logger)
                self.nodes.append(node)
                self.resource_listener.on_added(source=self, provider=self, resource=node)

        elif rtype == Constants.RES_TYPE_NETWORK.lower():
            network_count = resource.get(Constants.RES_COUNT, 1)
            name_prefix = resource.get(Constants.RES_NAME_PREFIX)
            site = resource.get(Constants.RES_SITE)

            for i in range(network_count):
                name = f"{name_prefix}{i}"
                network = DummyNetwork(label=label, name=name, site=site, logger=self.logger)
                self.networks.append(network)
                self.resource_listener.on_added(source=self, provider=self, resource=network)

        elif rtype == Constants.RES_TYPE_SERVICE.lower():
            service_count = resource.get(Constants.RES_COUNT, 1)
            service_name_prefix = resource.get(Constants.RES_NAME_PREFIX)

            for n in range(0, service_count):
                service_name = f"{self.name}-{service_name_prefix}-{n}"
                service = DummyService(label=label, name=service_name, image=image,
                                       x=exposed_attribute_x, logger=self.logger)

                self._services.append(service)
                self.resource_listener.on_added(source=self, provider=self, resource=service)

    def do_create_resource(self, *, resource: dict):
        """
        Called by create_resource(self, *, resource: dict) if resource has no external dependencies
        @param resource: resource attributes
        """
        label = resource.get(Constants.LABEL)
        self.logger.info(f"Creating resource={resource} using {self.label}")

        rtype = resource.get(Constants.RES_TYPE)

        if rtype == Constants.RES_TYPE_NODE.lower():
            for node in self.nodes:
                node.create()
                self.resource_listener.on_created(source=self, provider=self, resource=node)
        elif rtype == Constants.RES_TYPE_NETWORK.lower():
            for net in self.networks:
                net.create()
                self.resource_listener.on_created(source=self, provider=self, resource=net)
        elif rtype == Constants.RES_TYPE_SERVICE.lower():
            temp = [service for service in self.services if service.label == label]
            for service in temp:
                service.create()
                self.resource_listener.on_created(source=self, provider=self, resource=service)

    def do_delete_resource(self, *, resource: dict):
        self.logger.info(f"Deleting resource={resource} using {self.label}")

        label = resource.get(Constants.LABEL)
        rtype = resource.get(Constants.RES_TYPE)
        name_prefix = resource.get(Constants.RES_NAME_PREFIX)

        if rtype == Constants.RES_TYPE_NODE.lower():
            node_count = resource.get(Constants.RES_COUNT, 1)
            for i in range(node_count):
                name = f"{name_prefix}{i}"
                node = DummyNode(label=label, name=name, site=None, image=None, flavor=None, logger=self.logger)
                node.delete()
                self.resource_listener.on_deleted(source=self, provider=self, resource=node)
        elif rtype == Constants.RES_TYPE_NETWORK.lower():
            network_count = resource.get(Constants.RES_COUNT, 1)
            for i in range(network_count):
                name = f"{name_prefix}{i}"
                network = DummyNetwork(label=label, name=name, site=None, logger=self.logger)
                network.delete()
                self.resource_listener.on_deleted(source=self, provider=self, resource=network)
        elif rtype == Constants.RES_TYPE_SERVICE.lower():
            service_count = resource.get(Constants.RES_COUNT, 1)
            image = resource.get(Constants.RES_IMAGE)
            for n in range(0, service_count):
                service_name = f"{self.name}-{name_prefix}-{n}"
                service = DummyService(label=label, name=service_name, image=image, logger=self.logger)
                service.delete()
                self.resource_listener.on_deleted(source=self, provider=self, resource=service)



class AmSCROTException(Exception):
    """Base class for other exceptions"""
    pass

class AnsibleException(AmSCROTException):
    pass

class ParseConfigException(AmSCROTException):
    pass


class ResourceTypeNotSupported(AmSCROTException):
    pass


class ResourceNotAvailable(AmSCROTException):
    pass


class ConfigTypeNotSupported(AmSCROTException):
    pass


class StateException(AmSCROTException):
    pass


class ProviderTypeNotSupported(AmSCROTException):
    pass


class StitchPortNotFound(AmSCROTException):
    pass


class ProviderException(AmSCROTException):
    pass


class ControllerException(AmSCROTException):
    def __init__(self, exceptions):
        self.exceptions = exceptions

        self.message = f"Number Of Exceptions={len(exceptions)}:["

        for ex in exceptions:
            self.message += "\nmsg=" + ''.join(str(ex).splitlines())

        self.message += "\n]"
        super().__init__(self.message)



from collections import namedtuple
from typing import Dict, List

from amscrot.exceptions import StitchPortNotFound, AmSCROTException
from amscrot.util.constants import Constants

MEMBER_OF = 'member-of'
STITCH_PORT = 'stitch-port'
GROUP = 'group'
PRODUCER_FOR = 'producer-for'
CONSUMER_FOR = 'consumer-for'
PEER = 'peer'


class BadStitchPortException(AmSCROTException):
    """Base class for other exceptions"""
    pass


class ProviderPolicy:
    def __init__(self, *, type, stitch_ports, groups):
        self.type = type
        self.stitch_ports = stitch_ports

        for group in groups:
            if CONSUMER_FOR not in group:
                group[CONSUMER_FOR] = []
            if PRODUCER_FOR not in group:
                group[PRODUCER_FOR] = []

        self.groups = groups

    def __str__(self) -> str:
        lst = ["stitch_ports=" + str(self.stitch_ports), "groups=" + str(self.groups)]
        return str(lst)

    def __repr__(self) -> str:
        return self.__str__()


DetailedStitchInfo = namedtuple("DetailedStitchInfo", "stitch_port producer consumer producer_group consumer_group")
StitchInfo = namedtuple("StitchInfo", "stitch_port producer consumer")


def parse_policy(policy, policy_details, fp_dict=None) -> Dict[str, ProviderPolicy]:
    for k, v in policy.items():
        stitch_ports = v[STITCH_PORT] if STITCH_PORT in v else []
        effective_stitch_ports = []

        for stitch_port in stitch_ports:
            if not stitch_port.get('name'):
                raise BadStitchPortException(f"stitch port must have a name", stitch_port)

            stitch_port[Constants.PROVIDER] = k

            if 'preference' not in stitch_port:
                stitch_port['preference'] = 0

            handled_by_remote = False

            if fp_dict and stitch_port.get('profile') in fp_dict and fp_dict[stitch_port['profile']]:
                handled_by_remote = True
                fp_list = fp_dict[stitch_port['profile']]

                for detail in fp_list:
                    effective_stitch_port = stitch_port.copy()
                    effective_stitch_port.update(vars(detail))
                    effective_stitch_ports.append(effective_stitch_port)

            if handled_by_remote:
                continue

            if stitch_port.get('profile') and k in policy_details and policy_details[k] and stitch_port['profile'] in \
                    policy_details[k]:
                port_detail = policy_details[k][stitch_port['profile']]
                if isinstance(port_detail, list):
                    for detail in port_detail:
                        if not detail.get('name') or detail.get('name') == stitch_port['name']:
                            effective_stitch_port = stitch_port.copy()
                            effective_stitch_port.update(detail)
                            effective_stitch_ports.append(effective_stitch_port)
                else:
                    if not port_detail.get('name') or port_detail.get('name') == stitch_port['name']:
                        effective_stitch_port = stitch_port  # TODO
                        effective_stitch_port.update(port_detail)
                        effective_stitch_ports.append(effective_stitch_port)
            else:
                effective_stitch_port = stitch_port  # TODO
                effective_stitch_ports.append(effective_stitch_port)

        groups = v[GROUP] if GROUP in v else []

        for g in groups:
            g[Constants.PROVIDER] = k

        policy[k] = ProviderPolicy(type=k, stitch_ports=effective_stitch_ports, groups=groups)

    return policy


def load_remote_policy() -> Dict[str, ProviderPolicy]:
    from fabrictestbed_extensions.fablib.fablib import fablib
    from types import SimpleNamespace

    facility_ports = fablib.get_facility_ports()
    fp_dict = {}

    for fp in facility_ports.topology.facilities.values():
        for iface in fp.interface_list:
            local_name = iface.labels.local_name if iface.labels and iface.labels.local_name else None
            device_name = iface.labels.device_name if iface.labels and iface.labels.device_name else None
            region = iface.labels.region if iface.labels and iface.labels.region else None
            vlan_range = iface.labels.vlan_range if iface.labels else []
            fp_ns = SimpleNamespace(site=fp.site, local_name=local_name, device_name=device_name,
                                    region=region, vlan_range=vlan_range)

            fp_list = fp_dict.get(fp.name)

            if not fp_list:
                fp_list = []
                fp_dict[fp.name] = fp_list

            fp_list.append(fp_ns)

    policy = fablib.get_stitching_policy()
    policy_details = get_facility_ports()
    return parse_policy(policy, policy_details, fp_dict)


def load_policy(*, policy_file=None, content=None, load_details=True) -> Dict[str, ProviderPolicy]:
    import os
    import yaml
    import json

    if content:
        policy = yaml.safe_load(content)
        return parse_policy(policy, dict())

    if not policy_file:
        policy_file = os.path.join(os.path.dirname(__file__), 'stitching_policy.json')

    with open(policy_file, 'r') as fp:
        if policy_file.endswith(".json"):
            policy = json.load(fp)
        else:
            policy = yaml.load(fp, Loader=yaml.FullLoader)

    if load_details:
        policy_details = get_facility_ports()
        return parse_policy(policy, policy_details)
    else:
        return parse_policy(policy, dict())


def get_facility_ports():
    import os
    import json

    port_file = os.path.join(os.path.dirname(__file__), "stitching_policy_details.json")
    with open(port_file, 'r') as fp:
        ports = json.load(fp)
    return ports


def find_stitch_port_for_providers(policy: Dict[str, ProviderPolicy], providers: List[str]) -> List[DetailedStitchInfo]:
    provider1 = providers[0]
    provider2 = providers[1]
    stitch_infos = []

    for g in policy[provider1].groups:
        if provider2 in g[CONSUMER_FOR]:  # provider1's group is a consumer, find provider2's groups that are producers
            for producer_group in policy[provider2].groups:
                if provider1 in producer_group[PRODUCER_FOR]:
                    if g['name'] != producer_group['name']:
                        continue

                    for stitch_port in policy[g['provider']].stitch_ports:
                        if g['name'] in stitch_port[MEMBER_OF]:
                            stitch_info = DetailedStitchInfo(stitch_port=stitch_port,
                                                             producer=producer_group[Constants.PROVIDER],
                                                             consumer=g[Constants.PROVIDER],
                                                             producer_group=producer_group,
                                                             consumer_group=g)

                            stitch_infos.append(stitch_info)

                    for stitch_port in policy[producer_group['provider']].stitch_ports:
                        if producer_group['name'] in stitch_port[MEMBER_OF]:
                            stitch_info = DetailedStitchInfo(stitch_port=stitch_port,
                                                             producer=producer_group[Constants.PROVIDER],
                                                             consumer=g[Constants.PROVIDER],
                                                             producer_group=producer_group,
                                                             consumer_group=g)

                            stitch_infos.append(stitch_info)

        if provider2 in g[PRODUCER_FOR]:   # provider2's group is a producer find provider2's groups that are consumers
            for consumer_group in policy[provider2].groups:
                if provider1 in consumer_group[CONSUMER_FOR]:
                    if g['name'] != consumer_group['name']:
                        continue

                    for stitch_port in policy[g['provider']].stitch_ports:
                        if g['name'] in stitch_port[MEMBER_OF]:
                            stitch_info = DetailedStitchInfo(stitch_port=stitch_port,
                                                             producer=g[Constants.PROVIDER],
                                                             consumer=consumer_group[Constants.PROVIDER],
                                                             producer_group=g,
                                                             consumer_group=consumer_group)
                            stitch_infos.append(stitch_info)

                    for stitch_port in policy[consumer_group['provider']].stitch_ports:
                        if consumer_group['name'] in stitch_port[MEMBER_OF]:
                            stitch_info = DetailedStitchInfo(stitch_port=stitch_port,
                                                             producer=g[Constants.PROVIDER],
                                                             consumer=consumer_group[Constants.PROVIDER],
                                                             producer_group=g,
                                                             consumer_group=consumer_group)
                            stitch_infos.append(stitch_info)

    stitch_infos.sort(key=lambda sinfo: sinfo.stitch_port['preference'], reverse=True)
    removed_duplicates_stitch_infos = []

    for si in stitch_infos:
        found = False
        for csi in removed_duplicates_stitch_infos:
            if si.consumer == csi.consumer and si.producer == csi.producer and si.stitch_port == csi.stitch_port:
                found = True
                break

        if not found:
            removed_duplicates_stitch_infos.append(si)

    return removed_duplicates_stitch_infos


def check_options(k, v, adict):
    temp = adict.get('option')

    if temp:
        assert isinstance(temp, dict), "option must be a dictionary"

        if temp.get(k):
            if isinstance(v, str) and v == temp.get(k):
                return True, {(k, v)}

            if isinstance(v, list):
                for v_pair in v:
                    aset = set(v_pair.items())

                    if isinstance(temp.get(k), list):
                        for t_pair in temp.get(k):
                            bset = set(t_pair.items())

                            if aset.issubset(bset):
                                return True, aset

    return False, None


def peer_stitch_ports(stitch_infos: List[DetailedStitchInfo]):
    stitch_ports = []

    for si in stitch_infos:
        stitch_ports.append(si.stitch_port)

    effective_stitch_infos = []
    for si in stitch_infos:
        stitch_port = si.stitch_port
        stitch_port_provider = stitch_port['provider']
        stitch_port_providers = list(sorted([si.producer, si.consumer]))

        for sp in stitch_ports:
            if sp['provider'] == stitch_port_provider \
                    or sp['provider'] not in stitch_port_providers:
                continue

            if sp['provider'] != stitch_port_providers[0]:
                continue

            if stitch_port['name'] != sp['name']:
                continue

            acopy = stitch_port.copy()
            acopy['peer'] = sp.copy()
            effective_stitch_info = DetailedStitchInfo(stitch_port=acopy,
                                                       producer=si.producer,
                                                       consumer=si.consumer,
                                                       producer_group=si.producer_group,
                                                       consumer_group=si.consumer_group)
            effective_stitch_infos.append(effective_stitch_info)

    return effective_stitch_infos


def find_stitch_port(*, policy: Dict[str, ProviderPolicy], providers: List[str], site=None,
                     profile=None, options=None) -> DetailedStitchInfo or None:
    from amscrot.util.utils import get_logger

    logger = get_logger()
    stitch_infos = find_stitch_port_for_providers(policy, providers)
    stitch_infos = peer_stitch_ports(stitch_infos)

    logger.info(f"Found {len(stitch_infos)} stitch ports")

    if options:
        logger.info(f"Got options {options}")

        for k, v in options.items():
            for stitch_info in stitch_infos:
                if k == 'port_name':
                    k = 'name'

                stitch_ports = [stitch_info.stitch_port]

                if stitch_info.stitch_port.get('peer'):
                    stitch_ports.append(stitch_info.stitch_port.get('peer'))

                for stitch_port in stitch_ports:
                    if stitch_port.get(k) == v:
                        logger.info(f"Using stitch port based on port: {k}={v} and providers={providers}:{stitch_info}")
                        return stitch_info

                    ret, aset = check_options(k, v, stitch_port)

                    if ret:
                        logger.info(
                            f"Using stitch port based on port: {aset} and providers={providers}:{stitch_info}")
                        return stitch_info

                for g in [stitch_info.consumer_group, stitch_info.producer_group]:
                    if k == 'group_name':
                        k = 'name'

                    if g.get(k) == v:
                        logger.info(
                            f"Using stitch port based on group: {k}={v} and providers={providers}:{stitch_info}")

                        return stitch_info

                    ret, aset = check_options(k, v, g)

                    if ret:
                        logger.info(
                            f"Using stitch port based on group: {aset} and providers={providers}:{stitch_info}")
                        return stitch_info

            logger.warning(f"No stitch port based on {k}={v}")

        logger.info(f"Done with options {options}")

    if profile:
        for stitch_info in stitch_infos:
            if profile == stitch_info.stitch_port.get(Constants.RES_PROFILE):
                logger.info(f"Using stitch port based on profile={profile} and providers={providers}:{stitch_info}")
                return stitch_info

    if site:
        for stitch_info in stitch_infos:
            if site == stitch_info.stitch_port.get('site'):
                logger.info(f"Using stitch port based on site={site} and providers={providers}:{stitch_info}")
                return stitch_info

        logger.warning(f"did not find a stitch port for site={site} and providers={providers}")

    stitch_info = stitch_infos[0] if stitch_infos else None

    if not stitch_info:
        raise StitchPortNotFound(f"did not find a stitch port for providers={providers}")

    if len(stitch_infos) > 1:
        logger.info(f"Using stitch port based on preference for providers={providers}:{stitch_info}")
    else:
        logger.info(f"Using stitch port for providers={providers}:{stitch_info}")

    return stitch_info


def find_profile(network, resources):
    profile = network.attributes.get(Constants.RES_PROFILE)

    # if not profile:
    #     for dep in [dep for dep in network.dependencies if dep.resource.is_network]:
    #         profile = dep.resource.attributes.get(Constants.RES_PROFILE)
    #
    #         if profile:
    #             break

    if not profile:
        for net in [resource for resource in resources if resource.is_network]:
            if [dep for dep in network.dependencies if dep.resource == net]:
                profile = net.attributes.get(Constants.RES_PROFILE)

                if profile:
                    break
    return profile


def find_site(network, resources):
    site = network.attributes.get(Constants.RES_SITE)

    if not site:
        for dep in network.dependencies:
            if dep.resource.is_node:
                site = dep.resource.attributes.get(Constants.RES_SITE)

                if site:
                    break

    if not site:
        for node in [resource for resource in resources if resource.is_node]:
            if [dep for dep in node.dependencies if dep.resource == network]:
                site = node.attributes.get(Constants.RES_SITE)

                if site:
                    break
    return site


def clean_up_port(stitch_port):
    attrs = ["preference", "member-of"]

    for attr in attrs:
        stitch_port.pop(attr, None)

    peer = stitch_port.get('peer', {})
    attrs = ["preference", "member-of", "name"]

    for attr in attrs:
        peer.pop(attr, None)


def handle_stitch_info(config, policy, resources):
    from amscrot.util.utils import get_logger

    logger = get_logger()
    has_stitch_with = False

    for network in [resource for resource in resources if resource.is_network]:
        if Constants.RES_STITCH_INFO not in network.attributes:
            network.attributes[Constants.RES_STITCH_INFO] = []

        if Constants.RES_STITCH_INTERFACE not in network.attributes:
            network.attributes[Constants.RES_STITCH_INTERFACE] = []

    for network in [resource for resource in resources if resource.is_network]:
        if Constants.NETWORK_STITCH_WITH in network.attributes:
            from amscrot.util.config_models import DependencyInfo

            has_stitch_with = True
            dependency_info_dicts = network.attributes[Constants.NETWORK_STITCH_WITH]

            if not isinstance(dependency_info_dicts, list):
                option = network.attributes.get(Constants.NETWORK_STITCH_OPTION)
                dependency_info_dicts = [dict(network=dependency_info_dicts, stitch_option=option)]

            from amscrot.util.config_models import DependencyInfo

            for dependency_info_dict in dependency_info_dicts:
                assert("network" in dependency_info_dict)
                assert(isinstance(dependency_info_dict["network"], DependencyInfo))

            temp = set()

            for dep in network.dependencies:
                if dep.key == 'stitch_with.network':
                    # Dependency(key='stitch_with.network', resource=cnet@network, attribute='', is_external=True
                    from amscrot.util.config_models import Dependency
                    new_dep = Dependency(key=Constants.NETWORK_STITCH_WITH,
                                         resource=dep.resource, attribute=dep.attribute, is_external=True)
                    temp.add(new_dep)
                else:
                    temp.add(dep)

            network._resource_dependencies = temp

            for dependency_info_dict in dependency_info_dicts:
                dependency_info: DependencyInfo = dependency_info_dict['network']
                network_dependency = None

                for ed in network.dependencies:
                    if ed.key == Constants.NETWORK_STITCH_WITH and ed.resource.label == dependency_info.resource.label:
                        network_dependency = ed
                        break

                assert network_dependency is not None, "should never happen"
                assert network_dependency.is_external, "only network stitching across providers is supported"
                other_network = network_dependency.resource
                assert other_network.is_network, "only network stitching is supported"

                stitch_config = None
                option = dependency_info_dict.get(Constants.NETWORK_STITCH_OPTION)

                if option:
                    stitch_config = option.get(Constants.NETWORK_STITCH_CONFIG)

                if not stitch_config:
                    site = find_site(network, resources)
                    profile = find_profile(network, resources)
                    stitch_info = find_stitch_port(policy=policy,
                                                   providers=[network.provider.type, other_network.provider.type],
                                                   site=site,
                                                   profile=profile,
                                                   options=option)

                    clean_up_port(stitch_info.stitch_port)
                    stitch_info = StitchInfo(consumer=stitch_info.consumer,
                                             producer=stitch_info.producer,
                                             stitch_port=stitch_info.stitch_port)
                else:
                    logger.info(f"using supplied {Constants.NETWORK_STITCH_CONFIG}:{stitch_config.attributes}")
                    stitch_info = StitchInfo(consumer=stitch_config.attributes['consumer'],
                                             producer=stitch_config.attributes['producer'],
                                             stitch_port=stitch_config.attributes['stitch_port'])
                from amscrot.util.config_models import DependencyInfo

                if network.provider.type == stitch_info.consumer:
                    network.attributes[Constants.RES_STITCH_INTERFACE].append(DependencyInfo(resource=other_network,
                                                                                             attribute=''))
                else:
                    other_network.attributes[Constants.RES_STITCH_INTERFACE].append(DependencyInfo(resource=network,
                                                                                                   attribute=''))

                peer1 = {}
                peer2 = {}

                peer1.update(stitch_info.stitch_port['peer'])
                peer2.update(stitch_info.stitch_port)
                peer2.pop('peer')

                if network.provider.type != peer1['provider']:
                    pass

                stitch_info = StitchInfo(stitch_port=dict(),
                                         producer=stitch_info.producer, consumer=stitch_info.consumer)

                if network.provider.type == peer1['provider']:
                    stitch_info.stitch_port.update(peer1)
                    stitch_info.stitch_port['peer'] = dict()
                    stitch_info.stitch_port['peer'].update(peer2)
                else:
                    stitch_info.stitch_port.update(peer2)
                    stitch_info.stitch_port['peer'] = dict()
                    stitch_info.stitch_port['peer'].update(peer1)

                # Handle stitch port labels. The labels are inserted here to match it to a peering config
                stitch_info.stitch_port[Constants.LABELS] = sorted([network.label, other_network.label])
                network.attributes[Constants.RES_STITCH_INFO].append(stitch_info)
                stitch_info = StitchInfo(stitch_port=dict(),
                                         producer=stitch_info.producer, consumer=stitch_info.consumer)

                if other_network.provider.type == peer1['provider']:
                    stitch_info.stitch_port.update(peer1)
                    stitch_info.stitch_port['peer'] = dict()
                    stitch_info.stitch_port['peer'].update(peer2)
                else:
                    stitch_info.stitch_port.update(peer2)
                    stitch_info.stitch_port['peer'] = dict()
                    stitch_info.stitch_port['peer'].update(peer1)
                # Handle stitch port labels. The labels are inserted here to match it to a peering config
                stitch_info.stitch_port[Constants.LABELS] = sorted([network.label, other_network.label])
                other_network.attributes[Constants.RES_STITCH_INFO].append(stitch_info)

    for network in [resource for resource in resources if resource.is_network]:
        network.attributes.pop(Constants.NETWORK_STITCH_WITH, None)
        network.attributes.pop(Constants.NETWORK_STITCH_OPTION, None)

    if has_stitch_with:
        for resource in resources:
            resource.dependencies.clear()

        from amscrot.util.resource_dependency_helper import ResourceDependencyEvaluator, order_resources

        dependency_evaluator = ResourceDependencyEvaluator(resources, config.get_provider_configs())
        dependency_map = dependency_evaluator.evaluate()
        resources = order_resources(dependency_map)

    return resources


def fix_node_site(resource, resources):
    site = resource.attributes.get(Constants.RES_SITE)

    if site:
        return

    for dep in resource.dependencies:
        if dep.resource.is_network and dep.resource.provider == resource.provider:
            net = dep.resource
            site = net.attributes.get(Constants.RES_SITE)

            if site:
                resource.attributes[Constants.RES_SITE] = site
                return

            stitch_port = get_stitch_port_for_provider(resource=net.attributes, provider=net.provider.type)

            if isinstance(stitch_port, list) and len(stitch_port) > 1:
                return

            if isinstance(stitch_port, list) and len(stitch_port) == 1:
                stitch_port = stitch_port[0]

            if stitch_port:
                site = stitch_port.get(Constants.RES_SITE)  # TODO FIX PYCHARM WARNING
                resource.attributes[Constants.RES_SITE] = site
                return

    for net in [r for r in resources if r.is_network]:
        if [dep for dep in net.dependencies if net.provider == resource.provider and dep.resource == resource]:
            site = net.attributes.get(Constants.RES_SITE)

            if site:
                resource.attributes[Constants.RES_SITE] = site
                return

            stitch_port = get_stitch_port_for_provider(resource=net.attributes, provider=net.provider.type)

            if isinstance(stitch_port, list) and len(stitch_port) > 1:
                return

            if isinstance(stitch_port, list) and len(stitch_port) == 1:
                stitch_port = stitch_port[0]

            if stitch_port:
                site = stitch_port.get(Constants.RES_SITE)  # TODO FIX PYCHARM WARNING
                resource.attributes[Constants.RES_SITE] = site
                return


def fix_network_site(resource):
    site = resource.attributes.get(Constants.RES_SITE)

    if site:
        return

    stitch_ports = get_stitch_port_for_provider(resource=resource.attributes, provider=resource.provider.type)

    if stitch_ports is None:
        return

    if isinstance(stitch_ports, list) and len(stitch_ports) > 1:
        return

    if isinstance(stitch_ports, list) and len(stitch_ports) == 1:
        stitch_port = stitch_ports[0]
        site = stitch_port.get(Constants.RES_SITE)
        resource.attributes[Constants.RES_SITE] = site

    if isinstance(stitch_ports, dict):
        stitch_port = stitch_ports
        site = stitch_port.get(Constants.RES_SITE)
        resource.attributes[Constants.RES_SITE] = site


def get_vlan_from_range(*, resource: dict):
    vlan_range = get_vlan_range(resource=resource)

    if not vlan_range:
        return -1

    vlan_range = vlan_range[0]
    x = vlan_range.split("-")

    import random

    vlan = random.randrange(int(x[0]), int(x[1]) + 1)
    return vlan


def get_vlan_range(*, resource: dict):
    stitch_infos = resource.get(Constants.RES_STITCH_INFO)

    if not stitch_infos:
        return None

    if isinstance(stitch_infos, dict):  # This is for testing purposes.
        producer = stitch_infos['producer']
        consumer = stitch_infos['consumer']
        stitch_info = StitchInfo(stitch_port=stitch_infos['stitch_port'], producer=producer, consumer=consumer)
        stitch_infos = resource[Constants.RES_STITCH_INFO] = [stitch_info]

    stitch_ports = [stitch_infos[0].stitch_port]

    if PEER in stitch_infos[0].stitch_port:
        stitch_ports.append(stitch_infos[0].stitch_port[PEER])

    for stitch_port in stitch_ports:
        if Constants.STITCH_VLAN_RANGE in stitch_port:
            return stitch_port[Constants.STITCH_VLAN_RANGE]

    return None


def get_stitch_port_for_provider(*, resource: dict, provider: str):
    stitch_infos = resource.get(Constants.RES_STITCH_INFO)

    if not stitch_infos:
        return None

    if isinstance(stitch_infos, dict):  # This is for testing purposes.
        producer = stitch_infos['producer']
        consumer = stitch_infos['consumer']
        stitch_info = StitchInfo(stitch_port=stitch_infos['stitch_port'], producer=producer, consumer=consumer)
        stitch_infos = resource[Constants.RES_STITCH_INFO] = [stitch_info]

    stitch_ports = []

    for stitch_info in stitch_infos:
        temp_stitch_ports = [stitch_info.stitch_port]

        if PEER in stitch_info.stitch_port:
            temp_stitch_ports.append(stitch_info.stitch_port[PEER])

        stitch_port = next(filter(lambda sp: sp['provider'] == provider, temp_stitch_ports), None)

        if stitch_port is not None:
            stitch_ports.append(stitch_port)

    if len(stitch_ports) == 0:
        return None

    if len(stitch_ports) == 1:
        return stitch_ports[0]

    return stitch_ports




from amscrot.util.parser import Parser
from amscrot.util.config_models import DependencyInfo
from amscrot.exceptions import ParseConfigException
from amscrot.exceptions import ResourceTypeNotSupported
import pytest


def test_malformed_node():
    yaml_str = '''
resource:
  - node:
      - my_node:
          - provider: '{{ fabric.my_provider }}'
            count: 2
          - extra:
            count: 2
provider:
  - fabric:
    - my_provider:
       - user: user1
    '''

    with pytest.raises(ParseConfigException):
        Parser.parse(content=yaml_str)

def test_resources():
    yaml_str = '''
resource:
  - node:
      - my_node:
          - provider: '{{ fabric.my_provider }}'
            count: 2
      - my_node2:
          - provider: '{{ fabric.my_provider }}'
            count: 2
  - node:
      - my_node3:
          - provider: '{{ fabric.my_provider }}'
            count: 2
      - my_node4:
          - provider: '{{ fabric.my_provider }}'
            count: 2
provider:
  - fabric:
    - my_provider:
       - user: user1
    - my_other_provider:
       - user: user1
    '''

    providers, ordered_resources = Parser.parse(content=yaml_str)
    assert len(providers) == 2
    assert len(ordered_resources) == 4


def test_no_nodes_and_no_networks():
    yaml_str = '''
resource:
  - no_type:
      - my_no_type:
          - provider: '{{ fabric.my_provider }}'
provider:
  - fabric:
    - my_provider:
       - user: user1
    '''

    with pytest.raises(ResourceTypeNotSupported):
        Parser.parse(content=yaml_str)


def test_not_pointing_to_provider():
    yaml_str = '''
resource:
  - node:
      - my_node:
            - count: 2
provider:
  - fabric:
    - my_provider:
       - user: user1
    '''

    with pytest.raises(ParseConfigException):
        Parser.parse(content=yaml_str)


def test_no_provider_node():
    yaml_str = '''
resource:
  - node:
      - my_node:
            - no_provider: '{{ fabric.my_provider }}'
            
provider:
  - fabric:
    - my_provider:
       - user: user1
    '''

    with pytest.raises(ParseConfigException):
        Parser.parse(content=yaml_str)


def test_duplicate_providers():
    yaml_str = '''
provider:
  - fabric:
    - my_provider:
       - user: user1
  - fabric:
    - my_provider:
       - user: user2
    '''

    with pytest.raises(ParseConfigException):
        Parser.parse(content=yaml_str)


def test_duplicate_nodes():
    yaml_str = '''
resource:
  - node:
      - my_node:
          - provider: '{{ fabric.my_provider }}'
            count: 2
  - node:
      - my_node:
          - provider: '{{ fabric.my_provider }}'
            count: 5
provider:
  - fabric:
    - my_provider:
       - user: user1
    '''

    with pytest.raises(ParseConfigException):
        Parser.parse(content=yaml_str)


def test_duplicate_node_across_providers():
    yaml_str = '''
resource:
  - node:
      - my_node:
          - provider: '{{ fabric.my_provider1 }}'
            count: 2
  - node:
      - my_node:
          - provider: '{{ chi.my_provider2 }}' 
          
provider:
  - fabric:
    - my_provider1:
       - user: user1
  - chi:
    - my_provider2:
       - user: user1
    '''

    with pytest.raises(ParseConfigException):
        Parser.parse(content=yaml_str)


def test_duplicate_networks_across_providers():
    yaml_str = '''
resource:
  - network:
      - my_net:
          - provider: '{{ fabric.my_provider }}'
            count: 2
  - network:
      - my_net:
          - provider: '{{ chi.my_provider2 }}'

provider:
  - fabric:
    - my_provider:
       - user: user1
  - chi:
    - my_provider2:
       - user: user2
    '''

    with pytest.raises(ParseConfigException):
        Parser.parse(content=yaml_str)


def test_duplicate_across_types():
    yaml_str = '''
resource:
  - node:
      - my_var:
          - provider: '{{ fabric.my_var }}'
            count: 2
  - network:
      - my_var:
          - provider: '{{ chi.my_provider2 }}'
provider:
  - fabric:
    - my_var:
       - user: user1
  - chi:
    - my_provider2:
       - user: user1
    '''

    Parser.parse(content=yaml_str)


def test_different_types():
    yaml_str = '''
resource:
  - node:
      - my_node:
          - provider: '{{ fabric.my_provider }}'
            count: 1
            up: True
            vlans: [1, 2]
            network:
              type: ipv4
              ip:
                - interface: eth0
provider:
  - fabric:
    - my_provider:
       - user: user1
        '''
    _, resources = Parser.parse(content=yaml_str)
    assert len(resources) == 1
    assert len(resources[0].dependencies) == 0


def test_missing_variable():
    yaml_str = '''
provider:
  - fabric:
    - my_provider:
       - user: '{{ var.user_name }}'
resource:
  - node:
      - my_node:
          - provider: '{{ fabric.my_provider }}'
    '''
    with pytest.raises(ParseConfigException):
        Parser.parse(content=yaml_str)


def test_duplicate_variables():
    yaml_str = '''
variable:
  - slice_name:
      - default: "test_slice"
  - slice_name:
      - default: "user10"
provider:
  - fabric:
    - my_provider:
       - user: "some_user"
resource:
  - node:
      - my_node:
          - slice:  '{{ slice.my_slice }}'
  - slice:
      - my_slice:
          - provider: '{{ fabric.my_provider }}'
    '''
    with pytest.raises(ParseConfigException):
        Parser.parse(content=yaml_str)


def test_variables():
    yaml_str = '''
variable:
  - provider_name:
      - default: "test_provider"
  - user_name:
      - default: "user10"
  - open_ports:
      - default: [22, 443]
  - groups:
      - default:
  - vlan:

provider:
  - fabric:
    - my_provider:
       - user: '{{ var.user_name }}'
         name: '{{ var.provider_name }}'

resource:
  - node:
      - my_node:
          - provider: '{{ fabric.my_provider }}'
            ports:  '{{ var.open_ports }}'
            vlan:   '{{ var.vlan }}'
            groups: '{{ var.groups }}'
    '''
    with pytest.raises(ParseConfigException):
        Parser.parse(content=yaml_str)

    # now let's inject some values for groups and vlan
    var_dict = dict(groups=['g1'], vlan=5)

    providers, resources = Parser.parse(content=yaml_str, var_dict=var_dict)
    assert providers[0].name == "test_provider"
    assert providers[0].attributes['user'] == 'user10'
    assert resources[0].attributes['ports'] == [22, 443]
    assert resources[0].attributes['vlan'] == 5
    assert resources[0].attributes['groups'] == ['g1']


def test_dependencies():
    yaml_str = '''
resource:
  - node:
      - my_node:
          - provider: '{{ fabric.my_provider }}'
            simple_attr: '{{ network.my_network.vlan }}'
            list_attr:  [ '{{ network.my_network.ip }}', '{{ network.my_network.port }}']
            complex_attr:
              type: '{{ network.my_network.subnet }}'
              ip:
                - interface: '{{ network.my_network.os_interface.mask }}'
  - network:
      - my_network:
          - provider: '{{ fabric.my_provider }}'
provider:
  - fabric:
    - my_provider:
       - user: user1
    '''
    providers, resources = Parser.parse(content=yaml_str)
    assert len(providers) == 1
    assert len(resources) == 2

    for resource in resources:
        if resource.is_node:
            assert len(resource.attributes) == 3
            assert list(resource.attributes.keys()) == ['simple_attr', 'list_attr', 'complex_attr']
            assert isinstance(resource.attributes['simple_attr'], DependencyInfo)
            assert len(resource.attributes['list_attr']) == 2
            assert len(resource.attributes['complex_attr']) == 2
            assert len(resource.dependencies) == 5
            keys = set()
            attributes = set()

            for dep in resource.dependencies:
                keys.add(dep.key)
                attributes.add(dep.attribute)

            assert keys == {'complex_attr.ip.interface', 'list_attr', 'simple_attr', 'complex_attr.type'}
            assert attributes == {'ip', 'port', 'vlan', 'os_interface.mask', 'subnet'}

            temp = list()
            temp.append(resource.attributes['simple_attr'])
            temp.extend(resource.attributes['list_attr'])
            adict = resource.attributes['complex_attr']
            temp.append(adict['type'])
            ip_dep = adict['ip'][0]
            temp.append(ip_dep['interface'])
            assert len(resource.dependencies) == len(temp)
            temp = [d for d in temp if not isinstance(d, DependencyInfo)]
            assert not temp



from sense.client.requestwrapper import RequestWrapper

SENSE_CLIENT = None


def init_client(config):
    global SENSE_CLIENT

    if not SENSE_CLIENT:
        SENSE_CLIENT = SenseClient(config)


def get_client():
    return SENSE_CLIENT


# TODO Fix this and make sure the config is coming from amscrot credentials
class SenseClient(RequestWrapper):
    def __init__(self, config=None):
        self.temp_config = config
        super().__init__()

    def getConfig(self, config_file=None):
        if not self.temp_config:
            super().getConfig()
            return

        normalized_config = {}
        for k, v in self.temp_config.items():
            if k == 'verify':
                normalized_config['verify'] = v
            else:
                normalized_config[k.upper()] = v
        self.config = normalized_config
        self._validateConfig()
        self._setDefaults()



import logging

from amscrot.controller.controller import Controller
from amscrot.util import state as sutil
from amscrot.util import utils
from amscrot.util.config import WorkflowConfig
from typing import List
from amscrot.controller.provider_factory import default_provider_factory
from amscrot.exceptions import ControllerException
from amscrot.provider.dummy.dummy_provider import DummyService

from amscrot.model.state import ProviderState


class DummyFailCreateException(Exception):
    """Base class for other exceptions"""
    pass


def get_stats(*, states: List[ProviderState]):
    return utils.get_counters(states=states)


def cleanup_test_state(session: str):
    sutil.destroy_session(session)
    default_provider_factory._providers.clear()


def run_plan_workflow(*, session, config_str):
    config = WorkflowConfig.parse(content=config_str)

    # logger = utils.init_logger()
    logger = logging.getLogger(__name__)
    controller = Controller(config=config, logger=logger)
    states = sutil.load_states(session)
    controller.init(session=session, provider_factory=default_provider_factory, provider_states=states)
    controller.plan(provider_states=states)
    cr, dl = sutil.dump_plan(resources=controller.resources, to_json=False, summary=False)
    return cr, dl


def run_apply_workflow(*, session, config_str) -> List[ProviderState]:
    config = WorkflowConfig.parse(content=config_str)

    # logger = utils.init_logger()
    logger = logging.getLogger(__name__)
    states = sutil.load_states(session)
    controller = Controller(config=config, logger=logger)
    controller.init(session=session, provider_factory=default_provider_factory, provider_states=states)
    controller.plan(provider_states=[])
    controller.add(provider_states=[])

    try:
        controller.apply(provider_states=[])
    except ControllerException as e:
        assert isinstance(e.exceptions[0], DummyFailCreateException)

    states = controller.get_states()
    sutil.save_states(states, session)
    return states


def run_destroy_workflow(*, session, config_str) -> List[ProviderState]:
    config = WorkflowConfig.parse(content=config_str)
    logger = logging.getLogger(__name__)
    controller = Controller(config=config, logger=logger)
    states = sutil.load_states(session)
    controller.init(session=session, provider_factory=default_provider_factory, provider_states=states)
    controller.destroy(provider_states=states)
    sutil.save_states(states, session)

    if not states:
        sutil.destroy_session(session)

    return states


def test_simple_workflow():
    config_str = '''
provider:
  - dummy:
    - my_provider:
       - url: https://some_url:5000
resource:
  - service:
      - dtn:
         - provider: '{{ dummy.my_provider }}'
           image: ubuntu
    '''
    session = "test_simple"
    cleanup_test_state(session)
    cr, dl = run_plan_workflow(session=session, config_str=config_str)
    assert cr == 1 and dl == 0
    states = run_apply_workflow(session=session, config_str=config_str)
    assert len(states) == 1
    assert get_stats(states=states) == (0, 0, 1, 0, 0)
    states = run_destroy_workflow(session=session, config_str=config_str)
    assert len(states) == 0


def test_plan_workflow():
    config_str = '''
provider:
  - dummy:
    - my_provider:
       - url: https://some_url:5000
resource:
  - service:
      - dtn:
         - provider: '{{ dummy.my_provider }}'
           image: ubuntu
    '''

    config_str_2 = '''
provider:
  - dummy:
    - my_provider:
       - url: https://some_url:5000
resource:
  - service:
      - dtn2:
         - provider: '{{ dummy.my_provider }}'
           image: ubuntu
           count: 5
        '''

    config_str_3 = '''
provider:
  - dummy:
    - my_provider:
       - url: https://some_url:5000
resource:
  - service:
      - dtn2:
         - provider: '{{ dummy.my_provider }}'
           image: ubuntu
           count: 3
            '''
    session = "test_plan_workflow"
    cleanup_test_state(session)
    cr, dl = run_plan_workflow(session=session, config_str=config_str)
    assert (cr, dl) == (1, 0)
    states = run_apply_workflow(session=session, config_str=config_str)
    assert len(states) == 1
    assert get_stats(states=states) == (0, 0, 1, 0, 0)
    cr, dl = run_plan_workflow(session=session, config_str=config_str_2)
    assert (cr, dl) == (5, 1)

    states = run_apply_workflow(session=session, config_str=config_str_2)
    assert len(states) == 1
    assert get_stats(states=states) == (0, 0, 5, 0, 0)
    cr, dl = run_plan_workflow(session=session, config_str=config_str_3)
    assert (cr, dl) == (3, 5)

    states = run_destroy_workflow(session=session, config_str=config_str)
    assert len(states) == 0


def test_simple_workflow_with_create_failing():
    config_str = '''
provider:
  - dummy:
    - my_provider:
       - url: https://some_url:5000
resource:
  - service:
      - dtn:
         - provider: '{{ dummy.my_provider }}'
           image: ubuntu
    '''
    session = "test_simple_workflow_with_create_failing"
    cleanup_test_state(session)
    clazz = DummyService
    orig = clazz.create

    def create(self):
        raise DummyFailCreateException(f"Fail on purpose ... {self}")

    clazz.create = create
    states = run_apply_workflow(session=session, config_str=config_str)
    clazz.create = orig
    assert len(states) == 1
    assert get_stats(states=states) == (0, 0, 0, 0, 1)
    states = run_destroy_workflow(session=session, config_str=config_str)
    assert len(states) == 0


def test_service_dependency_workflow():
    config_str = '''
provider:
  - dummy:
    - my_provider:
       - url: https://some_url:5000
         name: prov1
  - dummy:
    - my_provider2:
       - url: https://some_other_url:5000
         name: prov2
resource:
  - service:
      - dtn1:
         - provider: '{{ dummy.my_provider }}'
           image: "centos"
           exposed_attribute_x: "{{ service.dtn2 }}"
  - service:
      - dtn2:
         - provider: '{{ dummy.my_provider2 }}'
           image: ubuntu
    '''
    session = "test_service_dependency"
    cleanup_test_state(session)
    states = run_apply_workflow(session=session, config_str=config_str)
    assert len(states) == 2
    assert get_stats(states=states) == (0, 0, 2, 0, 0)
    states = run_destroy_workflow(session=session, config_str=config_str)
    assert len(states) == 0


def test_service_dependency_workflow_with_pending():
    config_str = '''
provider:
  - dummy:
    - my_provider:
       - url: https://some_url:5000
         name: prov1
  - dummy:
    - my_provider2:
       - url: https://some_other_url:5000
         name: prov2
resource:
  - service:
      - dtn1:
         - provider: '{{ dummy.my_provider }}'
           image: "centos"
           exposed_attribute_x: "{{ service.dtn2 }}"
  - service:
      - dtn2:
         - provider: '{{ dummy.my_provider2 }}'
           image: ubuntu
    '''
    session = "test_service_dependency"
    cleanup_test_state(session)
    clazz = DummyService
    orig = clazz.create

    def create(self):
        if self.image == "ubuntu":
            raise DummyFailCreateException(f"Fail on purpose ... {self.image}")

    clazz.create = create
    states = run_apply_workflow(session=session, config_str=config_str)
    clazz.create = orig
    assert len(states) == 2
    assert get_stats(states=states) == (0, 0, 0, 1, 1)
    states = run_destroy_workflow(session=session, config_str=config_str)
    assert len(states) == 0


def test_attribute_dependency_workflow():
    config_str = '''
provider:
  - dummy:
    - my_provider:
       - url: https://some_url:5000
         name: prov1
  - dummy:
    - my_provider2:
       - url: https://some_other_url:5000
         name: prov2
resource:
  - service:
      - dtn1:
         - provider: '{{ dummy.my_provider }}'
           image: "centos"
           count: 5
           exposed_attribute_x: "{{ service.dtn2.exposed_attribute_x }}"
  - service:
      - dtn2:
         - provider: '{{ dummy.my_provider2 }}'
           image: ubuntu
    '''
    session = "test_attribute_dependency"
    cleanup_test_state(session)
    states = run_apply_workflow(session=session, config_str=config_str)
    assert len(states) == 2

    assert get_stats(states=states) == (0, 0, 6, 0, 0)
    x = states[0].service_states[0].attributes["exposed_attribute_x"]

    for state in states:
        for service_state in state.service_states:
            assert service_state.attributes["exposed_attribute_x"] == x

    states = run_destroy_workflow(session=session, config_str=config_str)
    assert len(states) == 0


def test_multiple_service_dependency_workflow():
    config_str = '''
provider:
  - dummy:
    - my_provider:
       - url: https://some_url:5000
         name: prov1
  - dummy:
    - my_provider2:
       - url: https://some_other_url:5000
         name: prov2
resource:
  - service:
      - dtn1:
         - provider: '{{ dummy.my_provider }}'
           image: "centos"
           exposed_attribute_x: ["{{ service.dtn2 }}", "{{ service.dtn3 }}"]
           count: 10
  - service:
      - dtn2:
         - provider: '{{ dummy.my_provider2 }}'
           image: ubuntu
           count: 2
  - service:
      - dtn3:
         - provider: '{{ dummy.my_provider2 }}'
           image: ubuntu
           count: 3
 
    '''
    session = "test_multiple_service_dependency"
    cleanup_test_state(session)
    states = run_apply_workflow(session=session, config_str=config_str)
    assert len(states) == 2
    assert get_stats(states=states) == (0, 0, 15, 0, 0)
    states = run_destroy_workflow(session=session, config_str=config_str)
    assert len(states) == 0



from typing import List, Optional, Dict, Any, Iterator, TYPE_CHECKING
from pydantic import BaseModel, Field
from amscrot.util.constants import Constants

if TYPE_CHECKING:
    from amscrot.model.metadata import Facility


class DiscoveredResource(BaseModel):
    """A single discovered resource item."""
    type: str
    data: Dict[str, Any] = Field(default_factory=dict)
    name: Optional[str] = None
    metadata: Optional[Any] = Field(default=None, exclude=True)
    """Typed metadata object (e.g. Facility) populated when native=False."""

    model_config = {"arbitrary_types_allowed": True}

    def model_post_init(self, __context: Any) -> None:
        # Auto-extract name from data if not explicitly set
        if self.name is None and 'name' in self.data:
            self.name = self.data['name']

    def to_dict(self) -> Dict[str, Any]:
        """Serialize back to the legacy dict format."""
        return {"type": self.type, "data": self.data}


class DiscoveryResult:
    """Container for discovered resources from a ServiceClient.

    Organizes resources by type and provides typed accessors for each
    resource category defined in Constants.
    """

    def __init__(self, items: Optional[List[DiscoveredResource]] = None):
        self._items: List[DiscoveredResource] = items or []

    # -- Typed accessors for named resource types --

    @property
    def network(self) -> List[DiscoveredResource]:
        return self.by_type(Constants.RES_NETWORK)

    @property
    def compute(self) -> List[DiscoveredResource]:
        return self.by_type(Constants.RES_COMPUTE)

    @property
    def storage(self) -> List[DiscoveredResource]:
        return self.by_type(Constants.RES_STORAGE)

    @property
    def allocation(self) -> List[DiscoveredResource]:
        return self.by_type(Constants.RES_ALLOCATION)

    @property
    def project(self) -> List[DiscoveredResource]:
        return self.by_type(Constants.RES_PROJECT)

    @property
    def facility(self) -> List[DiscoveredResource]:
        return self.by_type(Constants.RES_FACILITY)

    @property
    def facilities(self) -> List["Facility"]:
        """Return typed Facility objects from normalized (native=False) discovery."""
        return [
            item.metadata
            for item in self.by_type(Constants.RES_FACILITY)
            if item.metadata is not None
        ]

    @property
    def capability(self) -> List[DiscoveredResource]:
        return self.by_type(Constants.RES_CAPABILITY)

    @property
    def all(self) -> List[DiscoveredResource]:
        """Return all discovered resources."""
        return list(self._items)

    def by_type(self, type_str: str) -> List[DiscoveredResource]:
        """Filter resources by an arbitrary type string."""
        return [item for item in self._items if item.type == type_str]

    def summary(self) -> Dict[str, int]:
        """Return a dict of {type: count} for all resource types found."""
        counts: Dict[str, int] = {}
        for item in self._items:
            counts[item.type] = counts.get(item.type, 0) + 1
        return counts

    def to_list(self) -> List[Dict[str, Any]]:
        """Serialize to the legacy list-of-dicts format for backward compatibility."""
        return [item.to_dict() for item in self._items]

    def to_hierarchical(self) -> List[Dict[str, Any]]:
        """Return a hierarchical representation of normalized (native=False) Facility results.

        Each Facility is serialized with its nested compute, storage, network,
        and allocation resources.
        """
        from amscrot.model.metadata import Facility

        def _resource_dict(obj) -> Dict[str, Any]:
            """Generic typed-object -> dict using only the base + declared fields."""
            d: Dict[str, Any] = {}
            for field in type(obj).model_fields:
                val = getattr(obj, field, None)
                if val is not None:
                    d[field] = val
            return d

        def _fac_to_dict(fac: "Facility") -> Dict[str, Any]:
            return {
                "id": fac.id,
                "name": fac.name,
                "description": fac.description,
                "compute":     [_resource_dict(c) for c in (fac.compute     or [])],
                "storage":     [_resource_dict(s) for s in (fac.storage     or [])],
                "networks":    [_resource_dict(n) for n in (fac.networks    or [])],
                "allocations": [_resource_dict(a) for a in (fac.allocations or [])],
            }

        facilities = self.facilities  # List[Facility] from typed .metadata
        return [_fac_to_dict(f) for f in facilities]

    # -- Container protocol --

    def __len__(self) -> int:
        return len(self._items)

    def __bool__(self) -> bool:
        return len(self._items) > 0

    def __iter__(self) -> Iterator[DiscoveredResource]:
        return iter(self._items)

    def __repr__(self) -> str:
        return f"<DiscoveryResult items={len(self._items)} types={self.summary()}>"



#!/usr/bin/env python3
"""
Generate Python client from ESnet IRI OpenAPI specification.

This script can be used standalone or imported into CI/CD pipelines.

Credit: https://gitlab.com/amsc2/infrastructure-and-services/amsc-interfaces/amsc-api-python
"""

import argparse
import shutil
import subprocess
import sys
from pathlib import Path
from urllib.request import urlopen
from urllib.error import URLError


DEFAULT_API_URL = "https://iri-dev.ppg.es.net/openapi.json"
CONFIG_FILE = "config.yaml"
OUTPUT_DIR = "generated"


def check_api_accessibility(api_url: str) -> bool:
    """Check if the API is accessible."""
    try:
        with urlopen(api_url, timeout=5) as response:
            return response.status == 200
    except (URLError, Exception):
        return False


def check_tool_installed() -> bool:
    """Check if openapi-python-client is installed."""
    return shutil.which("openapi-python-client") is not None


def generate_client(api_url: str, config_file: str, output_dir: str, package_name: str) -> bool:
    """
    Generate the Python client.
    
    Returns:
        True if generation was successful, False otherwise.
    """
    script_dir = Path(__file__).parent
    config_path = script_dir / config_file
    output_path = script_dir / output_dir
    
    # Remove existing generated client
    if output_path.exists():
        print(f"[rm]  Removing existing generated client at {output_path}")
        shutil.rmtree(output_path)
    
    # Build command
    #cmd = [
    #    "openapi-python-client",
    #    "generate",
    #    "--url", api_url,
    #    "--config", str(config_path),
    #    "--output-path", str(output_path),
    #]
    
    cmd = [
        "openapi-generator",
        "generate",
        "-i", api_url,
        "-g", "python",
        "-o", str(output_path),
        "--package-name", str(package_name)
    ]
    
    print(f"[gen]  Generating Python client...")
    print(f"   Command: {' '.join(cmd)}")
    print()
    
    try:
        result = subprocess.run(
            cmd,
            cwd=script_dir,
            check=True,
            capture_output=False,
        )
        return result.returncode == 0
    except subprocess.CalledProcessError as e:
        print(f"[ERROR] Generation failed with exit code {e.returncode}")
        return False


def main():
    parser = argparse.ArgumentParser(
        description="Generate Python client from ESnet IRI OpenAPI specification"
    )
    parser.add_argument(
        "--api-url",
        default=DEFAULT_API_URL,
        help=f"OpenAPI JSON URL (default: {DEFAULT_API_URL})",
    )
    parser.add_argument(
        "--config",
        default=CONFIG_FILE,
        help=f"Config file path (default: {CONFIG_FILE})",
    )
    parser.add_argument(
        "--output",
        default=OUTPUT_DIR,
        help=f"Output directory (default: {OUTPUT_DIR})",
    )
    parser.add_argument(
        "--skip-check",
        action="store_true",
        help="Skip API accessibility check",
    )
    parser.add_argument(
        "--package-name",
        default="esnet_iri",
        help="Package name (default: esnet_iri)",
    )
    
    args = parser.parse_args()
    
    print("=" * 60)
    print("  ESnet IRI Python Client Generator")
    print("=" * 60)
    print()
    print(f"API URL: {args.api_url}")
    print(f"Config:  {args.config}")
    print(f"Output:  {args.output}")
    print(f"Package name:  {args.package_name}")
    print()
    
    # Check if tool is installed
    if not check_tool_installed():
        print("[ERROR] openapi-generator is not installed")
        print()
        print("To install (macOS), run:")
        print("  brew install openapi-generator")
        print()
        sys.exit(1)
    
    # Check API accessibility
    if not args.skip_check:
        print("[check] Checking API accessibility...")
        if check_api_accessibility(args.api_url):
            print("[OK] API is accessible")
            print()
        else:
            print(f"[ERROR] Cannot access API at {args.api_url}")
            print()
            print("Make sure the API server is running:")
            print("  make")
            print()
            print("Or use --skip-check to bypass this check")
            print()
            sys.exit(1)
    
    # Generate client
    if generate_client(args.api_url, args.config, args.output, args.package_name):
        print()
        print("=" * 60)
        print("  [OK] Client generated successfully!")
        print("=" * 60)
        print()
        print(f"Location: {args.output}")
        print()
        print("To install the client:")
        print(f"  cd {args.output}")
        print("  pip install -e .")
        print()
        print("To run tests:")
        print("  python tests/test_workflow_simple.py")
        print()
        sys.exit(0)
    else:
        print()
        print("[ERROR] Client generation failed")
        sys.exit(1)


if __name__ == "__main__":
    main()




import logging
import time

import ipaddress
from ipaddress import IPv4Network


from fabric_cf.orchestrator.orchestrator_proxy import Status
from fabrictestbed.slice_editor import Labels, Flags, Capacities

from fabrictestbed_extensions.fablib.slice import Slice
from fabrictestbed_extensions.fablib.facility_port import FacilityPort
from fabrictestbed_extensions.fablib.network_service import NetworkService, ServiceType
from typing import List

from fabrictestbed_extensions.fablib.interface import Interface

class Plugins:
    @staticmethod
    def load():
        n = NetworkService
        n.change_public_ip = Plugins.change_public_ip
        n.new_l3network = Plugins.new_l3network
        n.new_network_service = Plugins.new_network_service
        s = Slice
        s.add_facility_port = Plugins.add_facility_port
        s.add_l3network = Plugins.add_l3network
        f = FacilityPort
        f.new_facility_port = Plugins.new_facility_port

        
    
    def change_public_ip(self, ipv6: list[str] = None, ipv4: list[str] = None):
        labels = self.fim_network_service.labels
        if labels is None:
            labels = Labels()
        if self.fim_network_service.type == ServiceType.FABNetv4Ext:
            labels = Labels.update(labels, ipv4=ipv4)
            
        elif self.fim_network_service.type == ServiceType.FABNetv6Ext:
            labels = Labels.update(labels, ipv6=ipv6)
        
        self.fim_network_service.set_properties(labels=labels)
    
    def add_facility_port(self, name: str = None, site: str = None, labels: Labels = None, peer_labels: Labels = None, capacities: Capacities = None) -> NetworkService:
        """
        Adds a new L2 facility port to this slice
        :return: a new L2 facility port
        :rtype: NetworkService
        """
        return FacilityPort.new_facility_port(slice=self, name=name, site=site, labels=labels, peer_labels=peer_labels, capacities=capacities)

    @staticmethod
    def new_facility_port(
        slice: Slice = None, name: str = None, site: str = None,
        labels: Labels = None, peer_labels: Labels = None, capacities: Capacities = None):
        
        if capacities is None:
            capacities = Capacities(bw=10)
        
        fim_facility_port = slice.get_fim_topology().add_facility(name=name, site=site, capacities=capacities, labels=labels, peer_labels=peer_labels)
        
        return FacilityPort(slice, fim_facility_port)
    
    def add_l3network(
        self, name: str = None, interfaces: List[Interface] = [], type: str = "IPv4",
        technology: str = None) -> NetworkService:
        return NetworkService.new_l3network(
            slice=self, name=name, interfaces=interfaces, type=type, technology=technology
        )
    
    @staticmethod
    def new_l3network(
        slice: Slice = None,
        name: str = None,
        interfaces: List[Interface] = [],
        type: str = None, technology: str = None
    ):
        """
        Not inteded for API use. See slice.add_l3network
        """
        if type == "IPv6":
            nstype = ServiceType.FABNetv6
        elif type == "IPv4":
            nstype = ServiceType.FABNetv4
        elif type == "IPv4Ext":
            nstype = ServiceType.FABNetv4Ext
        elif type == "IPv6Ext":
            nstype = ServiceType.FABNetv6Ext
        elif type == "L3VPN":
            nstype = ServiceType.L3VPN
        else:
            raise Exception(
                "Invalid L3 Network Type: Allowed values [IPv4, IPv4Ext, IPv6, IPv6Ext, L3VPN]"
            )

        # TODO: need a fabnet version of this
        # validate nstype and interface List
        # NetworkService.validate_nstype(nstype, interfaces)

        return NetworkService.new_network_service(
            slice=slice, name=name, nstype=nstype, interfaces=interfaces, technology=technology
        )
    
    @staticmethod
    def new_network_service(
        slice: Slice = None,
        name: str = None,
        nstype: ServiceType = None,
        interfaces: List[Interface] = [],technology: str = None
    ):
        """
        Not inteded for API use. See slice.add_l2network


        Creates a new FABRIC network service and returns the fablib instance.

        :param slice: the fabric slice to build the network service with
        :type slice: Slice
        :param name: the name of the new network service
        :type name: str
        :param nstype: the type of network service to create
        :type nstype: ServiceType
        :param interfaces: a list of interfaces to
        :return: the new fablib network service
        :rtype: NetworkService
        """
        fim_interfaces = []
        for interface in interfaces:
            fim_interfaces.append(interface.get_fim_interface())

        logging.info(
            f"Create Network Service: Slice: {slice.get_name()}, Network Name: {name}, Type: {nstype}"
        )
        fim_network_service = slice.topology.add_network_service(
            name=name, nstype=nstype, interfaces=fim_interfaces, technology=technology
        )

        return NetworkService(slice=slice, fim_network_service=fim_network_service)  
